#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0621

from unittest.mock import MagicMock

import pytest

from bluetooth_mesh import Element
from bluetooth_mesh.messages.config import GATTNamespaceDescriptor
from bluetooth_mesh.messages.generic.onoff import GenericOnOffOpcode
from bluetooth_mesh.test.fixtures import *  # pylint: disable=W0614, W0401


class MockModel(MagicMock):
    MODEL_ID = (None, 0x1001)
    OPCODES = {GenericOnOffOpcode.ONOFF_STATUS}
    INSTANCES = []

    def __init__(self, *args, **kwargs):
        super().__init__(
            *[arg for arg in args if not isinstance(arg, Element)], **kwargs
        )
        MockModel.INSTANCES.append(self)


class MockVenforModel(MagicMock):
    MODEL_ID = (0x6666, 0x9999)
    OPCODES = {GenericOnOffOpcode.ONOFF_SET}
    INSTANCES = []

    def __init__(self, *args, **kwargs):
        super().__init__(
            *[arg for arg in args if not isinstance(arg, Element)], **kwargs
        )
        MockVenforModel.INSTANCES.append(self)


@pytest.fixture
def model_mocks():
    MockModel.INSTANCES = []
    MockVenforModel.INSTANCES = []
    return {MockModel, MockVenforModel}


@pytest.fixture
def element(model_mocks):
    class MockElement(Element):
        MODELS = model_mocks
        LOCATION = GATTNamespaceDescriptor.FORTY_SECOND

    mock_element = MockElement(MagicMock(), 0)
    return mock_element


def test_message_received(element, source, app_index, status_encoded, status_parsed):
    element.message_received(source, app_index, False, status_encoded)
    MockModel.INSTANCES[0].message_received.assert_called_once_with(
        source, app_index, False, status_parsed
    )


def test_other_opcode_message_received(element, source, app_index, get_encoded):
    element.message_received(source, app_index, False, get_encoded)
    MockModel.INSTANCES[0].message_received.assert_not_called()


def test_dev_message_received(
    element, source, net_index, status_encoded, status_parsed
):
    element.dev_key_message_received(source, True, net_index, status_encoded)
    MockModel.INSTANCES[0].dev_key_message_received.assert_called_once_with(
        source, True, net_index, status_parsed
    )


def test_other_opcode_dev_message_received(element, source, net_index, get_encoded):
    element.dev_key_message_received(source, True, net_index, get_encoded)
    MockModel.INSTANCES[0].dev_key_message_received.assert_not_called()


def test_update_model_configuration(element):
    config = dict()
    model_config = element.update_model_configuration(
        (MockModel.MODEL_ID[0], MockModel.MODEL_ID[1]), config
    )
    MockModel.INSTANCES[0].update_configuration.assert_called_once_with(model_config)


def test_update_vendor_model_configuration(element):
    config = dict()
    model_config = element.update_model_configuration(
        (MockVenforModel.MODEL_ID[0], MockVenforModel.MODEL_ID[1]), config
    )
    MockVenforModel.INSTANCES[0].update_configuration.assert_called_once_with(
        model_config
    )



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import construct
import pytest

from bluetooth_mesh.messages.health import *

valid = [
    # fmt: off
    pytest.param(
        HealthCurrentStatus,
        bytes.fromhex("043601"),
        {
            "test_id": 4,
            "company_id": 0x0136,
            "fault_array": [],
        },
        id='HealthCurrentStatus - without faults'
    ),
    pytest.param(
        HealthCurrentStatus,
        bytes.fromhex("043601020304"),
        {
            "test_id": 4,
            "company_id": 0x0136,
            "fault_array": [0x02, 0x03, 0x04],
        },
        id='HealthCurrentStatus - with faults'
    ),
    pytest.param(
        HealthFaultGet,
        bytes.fromhex("3601"),
        {
            "company_id": 0x0136,
        },
        id='HealthFaultGet'
    ),
    pytest.param(
        HealthFaultClear,
        bytes.fromhex("3601"),
        {
            "company_id": 0x0136,
        },
        id='HealthFaultClear'
    ),
    pytest.param(
        HealthFaultTest,
        bytes.fromhex("033601"),
        {
            "test_id": 3,
            "company_id": 0x0136,
        },
        id='HealthFaultTest'
    ),
    pytest.param(
        HealthFaultStatus,
        bytes.fromhex("043601"),
        {
            "test_id": 4,
            "company_id": 0x0136,
            "fault_array": [],
        },
        id='HealthCurrentStatus - without faults'
    ),
    pytest.param(
        HealthFaultStatus,
        bytes.fromhex("043601020304"),
        {
            "test_id": 4,
            "company_id": 0x0136,
            "fault_array": [0x02, 0x03, 0x04],
        },
        id='HealthCurrentStatus - with faults'
    ),
    pytest.param(
        HealthPeriodGet,
        bytes.fromhex(""),
        {},
        id='HealthPeriodGet',
    ),
    pytest.param(
        HealthPeriodSet,
        bytes.fromhex("0f"),
        {
            "fast_period_divisor": 15
        },
        id='HealthPeriodSet',
    ),
    pytest.param(
        HealthPeriodStatus,
        bytes.fromhex("06"),
        {
            "fast_period_divisor": 6
        },
        id='HealthPeriodStatus',
    ),
    pytest.param(
        HealthAttentionGet,
        bytes.fromhex(""),
        {},
        id='HealthAttentionGet',
    ),
    pytest.param(
        HealthAttentionSet,
        bytes.fromhex("07"),
        {
            "attention": 7
        },
        id='HealthAttentionSet',
    ),
    pytest.param(
        HealthAttentionStatus,
        bytes.fromhex("06"),
        {
            "attention": 6
        },
        id='HealthAttentionSet',
    ),
    # fmt: on
]

parse_invalid = [
    # fmt: off
    pytest.param(
        HealthPeriodSet,
        bytes.fromhex("10"),
        construct.core.ValidationError,
        id="HealthPeriodSet - wrong value"
    ),
    # fmt: on
]

build_invalid = [
    # fmt: off
    pytest.param(
        HealthPeriodSet,
        {
            "fast_period_divisor": 16
        },
        construct.core.ValidationError,
        id="HealthPeriodSet - wrong value"
    ),
    # fmt: on
]


@pytest.mark.parametrize("message,encoded,decoded", valid)
def test_build(message, encoded, decoded):
    result = message.build(obj=decoded)
    assert result == encoded


@pytest.mark.parametrize("message,encoded,decoded", valid)
def test_parse(message, encoded, decoded):
    result = message.parse(data=encoded)
    assert result == decoded


@pytest.mark.parametrize("message,encoded,exception", parse_invalid)
def test_parse_invalid(message, encoded, exception):
    with pytest.raises(exception):
        message.parse(data=encoded)


@pytest.mark.parametrize("message,decoded,exception", build_invalid)
def test_build_invalid(message, decoded, exception):
    with pytest.raises(exception):
        message.build(obj=decoded)


def test_build_health_message():
    data = HealthMessage.build(
        dict(opcode=HealthOpcode.ATTENTION_SET, params=dict(attention=6))
    )

    assert data == bytes.fromhex("800506")


def test_parse_health_message():
    msg = HealthMessage.parse(bytes.fromhex("800506"))

    assert msg == dict(opcode=HealthOpcode.ATTENTION_SET, params=dict(attention=6))


def test_parse_health_message_bad_opcode():
    with pytest.raises(construct.core.ValidationError):
        HealthMessage.parse(bytes.fromhex("8011"))



import logging
from uuid import UUID

from functools import lru_cache
import dbus_next.errors
import dbus_next.signature
import pytest

from bluetooth_mesh.application import Application, Element
from bluetooth_mesh.crypto import NetworkKey

pytestmark = [pytest.mark.asyncio]


class Node:
    def __init__(self, path, token, address):
        self.path = path
        self.configuration = []
        self.address = address
        self.token = token


@pytest.fixture
def node():
    return Node("/org/bluez/mesh/node3342c6c574ed4145888ce1d978a1f145", 0x4321, 0x100)


@pytest.fixture
def dbus(monkeypatch, node, application):
    logger = logging.getLogger("dbus")

    class MockInterface:
        def __init__(self, proxy_object, interface_name):
            self.proxy_object = proxy_object
            self.interface_name = interface_name

    class MockDbusInterface(MockInterface):
        def on_name_owner_changed(self, callback):
            pass

        def off_name_owner_changed(self, callback):
            pass

        async def call_get_name_owner(self, service_name):
            pass

    class MockNodeInterface(MockInterface):
        NODE = node

        async def get_addresses(self):
            return [self.NODE.address]

    class MockNetworkInterface(MockInterface):
        NODE = node

        async def call_attach(self, app_root, token):
            if self.NODE is None:
                raise dbus_next.errors.DBusError(
                    "org.bluez.mesh.Failure", "Node doesn't exist"
                )

            if self.NODE.token != token:
                raise dbus_next.errors.DBusError(
                    "org.bluez.mesh.Failure", "Invalid token"
                )

            return self.NODE.path, self.NODE.configuration

        async def call_import(
            self, app_root, uuid, dev_key, net_key, net_index, flags, iv_index, unicast
        ):
            path = f"/org/bluez/mesh/node{uuid.hex()}"

            if MockMessageBus.SERVICE_OBJECTS.get(("org.bluez.mesh", path)):
                raise dbus_next.errors.DBusError(
                    "org.bluez.mesh.Failure", "Node exists"
                )

            MockMessageBus.SERVICE_OBJECTS[
                ("org.bluez.mesh", path)
            ] = MockNodeProxyObject
            MockNodeInterface.NODE = MockNetworkInterface.NODE = Node(
                path, 0x5678, unicast
            )
            application.join_complete(0x5678)

    class MockManagementInterface(MockInterface):
        async def call_import_remote_node(self, primary, count, device_key):
            pass

    class MockPropertiesInterface(MockInterface):
        def on_properties_changed(self, callback):
            pass

    class MockProxyObject:
        INTERFACES = {}

        def __init__(self, service_name, object_path, introspection):
            self.service_name = service_name
            self.object_path = object_path
            self.introspection = introspection

        def get_interface(self, interface_name):
            logger.debug("get interface %s", interface_name)
            return self.INTERFACES[interface_name](self, interface_name)

    class MockDbusProxyObject(MockProxyObject):
        INTERFACES = {"org.freedesktop.DBus": MockDbusInterface}

    class MockNetworkProxyObject(MockProxyObject):
        INTERFACES = {"org.bluez.mesh.Network1": MockNetworkInterface}

    class MockNodeProxyObject(MockProxyObject):
        INTERFACES = {
            "org.freedesktop.DBus.Properties": MockPropertiesInterface,
            "org.bluez.mesh.Node1": MockNodeInterface,
            "org.bluez.mesh.Management1": MockManagementInterface,
        }

    class MockMessageBus:
        SERVICE_OBJECTS = {
            ("org.freedesktop.DBus", "/org/freedesktop/DBus"): MockDbusProxyObject,
            ("org.bluez.mesh", "/org/bluez/mesh"): MockNetworkProxyObject,
        }

        def __init__(self, bus_type):
            pass

        async def connect(self):
            return self

        def disconnect(self):
            pass

        async def introspect(self, service_name, object_path):
            pass

        def export(self, object_path, interface):
            pass

        def unexport(self, object_path):
            pass

        def get_proxy_object(self, service_name, object_path, introspection):
            logger.debug("get proxy object for %s:%s", service_name, object_path)

            return self.SERVICE_OBJECTS[(service_name, object_path)](
                service_name, object_path, introspection
            )

    if node:
        MockMessageBus.SERVICE_OBJECTS[
            ("org.bluez.mesh", node.path)
        ] = MockNodeProxyObject

    monkeypatch.setattr("dbus_next.aio.MessageBus", MockMessageBus)


@pytest.fixture
def application(event_loop):
    class MockElement(Element):
        LOCATION = 0

    class MockApplication(Application):
        ELEMENTS = {0: MockElement}

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.address = 0x200
            self.__uuid = UUID("3a1f9d4659b74b009dfd32cf9e758b7a")

        @property
        def uuid(self) -> UUID:
            return self.__uuid

        @uuid.setter
        def uuid(self, value):
            self.__uuid = value

        @property
        def primary_net_key(self):
            return 0, NetworkKey(bytes.fromhex("b37e0124a48a406bd2b092e82bea6aed"))

    return MockApplication(event_loop)


async def test_attach_bad_token(application, dbus, node):
    async with application:
        with pytest.raises(dbus_next.errors.DBusError, match="Invalid token"):
            await application.attach(node.token + 1)


async def test_attach_good_token(application, dbus, node):
    async with application:
        await application.attach(node.token)
        assert (
            application.node_interface._interface.proxy_object.object_path == node.path
        )


@pytest.mark.parametrize("node", [None])
class TestImportOnConnect:
    async def test_import(self, application, dbus):
        async with application:
            await application.connect()

            assert (
                application.node_interface._interface.proxy_object.object_path
                == f"/org/bluez/mesh/node{application.uuid.hex}"
            )
            assert application.address == 0x200

    async def test_missing_address(self, application, dbus):
        application.address = None

        async with application:
            with pytest.raises(
                AttributeError, match="Application didn't provide an address"
            ):
                await application.connect()


class TestAttachOnConnect:
    async def test_read_address_after_attach(self, application, dbus, node):
        application.address = None
        application.token_ring.token = node.token

        async with application:
            await application.connect()

            assert (
                application.node_interface._interface.proxy_object.object_path
                == node.path
            )
            assert application.address == node.address

            with pytest.raises(AttributeError, match="Can't set address"):
                application.address = 5

    async def test_import_new_on_attach_with_bad_token(self, application, dbus, node):
        application.token_ring.token = node.token + 1

        async with application:
            await application.connect()

            assert (
                application.node_interface._interface.proxy_object.object_path
                == f"/org/bluez/mesh/node{application.uuid.hex}"
            )
            assert application.address == 0x200

    async def test_import_existing_with_bad_token(self, application, dbus, node):
        application.token_ring.token = node.token + 1
        _, uuid = node.path.split("/node")
        application.uuid = UUID(uuid)

        async with application:
            with pytest.raises(dbus_next.errors.DBusError, match="Node exists"):
                await application.connect()



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
"""
This module implements D-Bus interfaces that need to be exposed by an
application using BlueZ mesh stack, described in detail in mesh-api.txt_.

They are not meant to be used directly. See :py:mod:`bluetooth_mesh.application` instead.

.. _mesh-api.txt: https://git.kernel.org/pub/scm/bluetooth/bluez.git/tree/doc/mesh-api.txt
"""
# pylint: disable=R0201,W0613,W0622

import asyncio
import logging
import os
import socket
from collections import defaultdict
from contextlib import suppress
from datetime import timedelta
from enum import Enum
from typing import Any, Dict, Mapping, Optional, Sequence, Tuple
from uuid import UUID

from dbus_next import DBusError, Variant
from dbus_next.message import MessageFlag
from dbus_next.service import PropertyAccess, ServiceInterface, dbus_property, method

from bluetooth_mesh.crypto import ApplicationKey, DeviceKey, NetworkKey
from bluetooth_mesh.utils import Signal


class DBusService:
    NAME = "org.freedesktop.DBus"
    PATH = "/org/freedesktop/DBus"


class DBusInterface:
    def __init__(self, dbus_service):
        self._interface = dbus_service.get_interface("org.freedesktop.DBus")

    async def get_name_owner(self, name):
        return await self._interface.call_get_name_owner(name)

    def on_name_owner_changed(self, callback):
        self._interface.on_name_owner_changed(callback)

    def off_name_owner_changed(self, callback):
        self._interface.off_name_owner_changed(callback)

    def on_name_lost(self, callback):
        self._interface.on_name_lost(callback)

    def off_name_lost(self, callback):
        self._interface.off_name_lost(callback)

    async def get_service(self, service):
        owner = asyncio.Future()

        def name_owner_changed(name, old_owner, new_owner):
            if name != service.NAME:
                return

            if not new_owner:
                return

            owner.set_result(new_owner)

        try:
            self.on_name_owner_changed(name_owner_changed)
            return await self.get_name_owner(service.NAME)
        except DBusError as ex:
            return await owner
        finally:
            self.off_name_owner_changed(name_owner_changed)


class MeshService:
    NAME = "org.bluez.mesh"
    PATH = "/org/bluez/mesh"


def extract_model_config(configuration):
    vendor_id = None
    model_conf_dict = {}

    for key, variant in configuration.items():
        if key == "Vendor":
            vendor_id = variant.value
            continue

        if key == "Bindings":
            model_conf_dict["bindings"] = variant.value
            continue

        if key == "PublicationPeriod":
            model_conf_dict["publication_period"] = timedelta(
                milliseconds=variant.value
            )
            continue

        if key == "Subscriptions":
            subs = []

            for addr in variant.value:
                if addr.signature == "q":
                    subs.append(addr.value)

                elif addr.signature == "ay":
                    subs.append(UUID(bytes=addr.value))

            model_conf_dict["subscriptions"] = set(subs)

    return vendor_id, model_conf_dict


class ElementInterface(ServiceInterface):
    def __init__(self, element):
        self.element = element
        super().__init__(name="org.bluez.mesh.Element1")

    @dbus_property(name="Models", access=PropertyAccess.READ)
    def get_models(self) -> "a(qa{sv})":
        models = []

        for model_id, publish, subscribe in self.element.models:
            models.append(
                [
                    model_id,
                    {
                        "Publish": Variant("b", publish),
                        "Subscribe": Variant("b", subscribe),
                    },
                ]
            )

        return models

    @dbus_property(name="VendorModels", access=PropertyAccess.READ)
    def get_vendor_models(self) -> "a(qqa{sv})":
        models = []

        for model_id, publish, subscribe in self.element.vendor_models:
            models.append(
                [
                    *model_id,
                    {
                        "Publish": Variant("b", publish),
                        "Subscribe": Variant("b", subscribe),
                    },
                ]
            )

        return models

    @method(name="MessageReceived")
    def message_received(
        self, source: "q", key_index: "q", destination: "v", data: "ay"
    ):
        if destination.signature == "q":
            destination = destination.value
        elif destination.signature == "ay":
            destination = UUID(destination.value)

        self.element.message_received(source, key_index, destination, data)

    @method(name="DevKeyMessageReceived")
    def dev_key_message_received(
        self, source: "q", remote: "b", net_index: "q", data: "ay"
    ):
        self.element.dev_key_message_received(source, remote, net_index, data)

    @method(name="UpdateModelConfiguration")
    def update_model_configuration(self, model_id: "q", configuration: "a{sv}"):
        vendor_id, model_config = extract_model_config(configuration)
        self.element.update_model_configuration((vendor_id, model_id), model_config)

    @dbus_property(name="Index", access=PropertyAccess.READ)
    def get_index(self) -> "y":
        return self.element.index

    @dbus_property(name="Location", access=PropertyAccess.READ)
    def get_location(self) -> "q":
        return self.element.location


class ProvisionAgentInterface(ServiceInterface):
    def __init__(self, application):
        self.application = application
        self.logger = logging.getLogger("ProvisionAgentInterface")
        super().__init__(name="org.bluez.mesh.ProvisionAgent1")

    @method(name="PrivateKey")
    def private_key(self) -> "ay":
        return self.application.private_key()

    @method(name="PublicKey")
    def public_key(self) -> "ay":
        return self.application.public_key()

    @method(name="DisplayString")
    def display_string(self, value: "s"):
        self.application.display_string(value)

    @method(name="DisplayNumeric")
    def display_numeric(self, type: "s", number: "u"):
        self.application.display_numeric(type, number)

    @method(name="PromptNumeric")
    def prompt_numeric(self, type: "s") -> "u":
        return self.application.prompt_numeric(type)

    @method(name="PromptStatic")
    def prompt_static(self, type: "s") -> "ay":
        return self.application.prompt_static(type)

    @method(name="Cancel")
    def cancel(self):
        pass

    @dbus_property(name="Capabilities", access=PropertyAccess.READ)
    def get_capabilities(self) -> "as":
        return [cap.value for cap in self.application.capabilities]

    @dbus_property(name="OutOfBandInfo", access=PropertyAccess.READ)
    def out_of_band_info(self) -> "as":
        return [oob.value for oob in self.application.oob_info]

    @dbus_property(name="URI", access=PropertyAccess.READ)
    def uri(self) -> "s":
        return self.application.uri


class ProvisionerInterface(ServiceInterface):
    def __init__(self, application):
        self.application = application
        self.logger = logging.getLogger("ProvisionerInterface")
        super().__init__(name="org.bluez.mesh.Provisioner1")

    @method(name="ScanResult")
    def scan_result(self, rssi: "n", data: "ay", options: "a{sv}"):
        self.application.scan_result(rssi, data, options)

    @method(name="RequestProvData")
    def request_prov_data(self, count: "y") -> "qq":
        return self.application.request_prov_data(count)

    @method(name="AddNodeComplete")
    def add_node_complete(self, uuid: "ay", unicast: "q", count: "y"):
        self.application.add_node_complete(uuid=uuid, unicast=unicast, count=count)

    @method(name="AddNodeFailed")
    def add_node_failed(self, uuid: "ay", reason: "s"):
        self.application.add_node_failed(uuid=uuid, reason=reason)


class _ApplicationInterface(ServiceInterface):
    def __init__(self, application):
        self.application = application
        self.logger = logging.getLogger("ApplicationInterface")
        super().__init__(name="org.bluez.mesh.Application1")

    @method(name="JoinComplete")
    async def join_complete(self, token: "t"):
        self.application.join_complete(token)

    @method(name="JoinFailed")
    async def join_failed(self, reason: "s"):
        self.application.join_failed(reason)

    @dbus_property(name="CompanyID", access=PropertyAccess.READ)
    def get_company_id(self) -> "q":
        return self.application.company_id

    @dbus_property(name="ProductID", access=PropertyAccess.READ)
    def get_product_id(self) -> "q":
        return self.application.product_id

    @dbus_property(name="VersionID", access=PropertyAccess.READ)
    def get_version_id(self) -> "q":
        return self.application.version_id


class _ApplicationInterfaceCRPL(_ApplicationInterface):
    @dbus_property(name="CRPL", access=PropertyAccess.READ)
    def get_crpl(self) -> "q":
        return self.application.crpl


def ApplicationInterface(application):
    if application.crpl is not None:
        return _ApplicationInterfaceCRPL(application)
    else:
        return _ApplicationInterface(application)


class NetworkInterface:
    def __init__(self, mesh_service):
        self._interface = mesh_service.get_interface("org.bluez.mesh.Network1")

    def _extract_model_config(self, configuration):
        configuration_dict = defaultdict(dict)

        for element, element_conf in configuration:
            for model_id, model_conf in element_conf:
                vendor_id, model_conf_dict = extract_model_config(model_conf)

                configuration_dict[element][(vendor_id, model_id)] = model_conf_dict

        return configuration_dict

    async def join(self, app_defined_root: str, uuid: UUID) -> None:
        await self._interface.call_join(app_defined_root, uuid.bytes)

    async def cancel(self) -> None:
        await self._interface.call_cancel()

    async def attach(
        self, app_defined_root: str, token: int
    ) -> Tuple[str, Dict[int, Dict[Tuple[Optional[int], int], Mapping[str, Any]]]]:
        path, configuration = await self._interface.call_attach(app_defined_root, token)

        return path, self._extract_model_config(configuration)

    async def attach_fd(
        self, app_defined_root: str, token: int
    ) -> Tuple[
        str,
        Dict[int, Dict[Tuple[Optional[int], int], Mapping[str, Any]]],
        socket.socket,
    ]:
        path, configuration, fd = await self._interface.call_attach_fd(
            app_defined_root, token
        )

        sock = socket.fromfd(fd, socket.AF_UNIX, socket.SOCK_DGRAM)
        sock.setblocking(False)

        return path, self._extract_model_config(configuration), sock

    async def attach_unix(
        self, app_defined_root: str, token: int, socket_path: str
    ) -> Tuple[
        str,
        Dict[int, Dict[Tuple[Optional[int], int], Mapping[str, Any]]],
        socket.socket,
    ]:
        with suppress(FileNotFoundError):
            os.unlink(socket_path)

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
        sock.setblocking(False)
        sock.bind(socket_path)

        path, configuration = await self._interface.call_attach_unix(
            app_defined_root, token, socket_path
        )

        return path, self._extract_model_config(configuration), sock

    async def leave(self, token: int) -> None:
        await self._interface.call_leave(token)

    async def create_network(self, app_root: str, uuid: UUID) -> None:
        return await self._interface.call_create_network(app_root, uuid.bytes)

    async def import_node(
        self,
        app_root: str,
        uuid: UUID,
        dev_key: DeviceKey,
        net_key: NetworkKey,
        net_index: int,
        flags: Mapping[str, Any],
        iv_index: int,
        unicast: int,
    ) -> None:

        await self._interface.call_import(
            app_root,
            uuid.bytes,
            dev_key.bytes,
            net_key.bytes,
            net_index,
            flags,
            iv_index,
            unicast,
        )


class AclInterface:
    def __init__(self, mesh_service):
        self._interface = mesh_service.get_interface(
            "org.bluez.mesh.AccessControlList1"
        )

    async def grant_access(
        self, uuid: UUID, dev_key: DeviceKey, net_key: NetworkKey
    ) -> int:
        return await self._interface.call_grant_access(uuid, dev_key, net_key)

    async def revoke_access(self, token: int):
        return await self._interface.call_revoke_access(token)


class IvState(Enum):
    INIT = "init"
    NORMAL = "normal"
    UPDATING = "updating"
    NORMAL_HOLD = "hold"


class NodeInterface:
    def __init__(self, node_service):
        self._interface = node_service.get_interface("org.bluez.mesh.Node1")
        self._properties = node_service.get_interface("org.freedesktop.DBus.Properties")
        self._properties.on_properties_changed(self._on_properties_changed)

    def _on_properties_changed(
        self, interface: str, changed: Mapping[str, Any], invalidated: Sequence[str]
    ):
        if interface != "org.bluez.mesh.Node1":
            return

        for name, variant in changed.items():
            val = variant.value

            if name == "IvIndex":
                sig = self.iv_index.signal
            elif name == "IvUpdate":
                sig = self.iv_update.signal
            elif name == "IvState":
                sig = self.iv_state.signal
                val = IvState(variant.value)
            else:
                sig = None

            if sig:
                asyncio.ensure_future(sig.emit(val))

    async def send(
        self, element_path: str, destination: int, key_index: int, data: bytes
    ) -> None:
        await self._interface.call_send(
            element_path,
            destination,
            key_index,
            data,
            flags=MessageFlag.NO_REPLY_EXPECTED,
        )

    async def dev_key_send(
        self,
        element_path: str,
        destination: int,
        remote: bool,
        net_index: int,
        data: bytes,
    ) -> None:
        await self._interface.call_dev_key_send(
            element_path,
            destination,
            remote,
            net_index,
            data,
            flags=MessageFlag.NO_REPLY_EXPECTED,
        )

    async def add_net_key(
        self,
        element_path: str,
        destination: int,
        subnet_index: int,
        net_index: int,
        update: bool,
    ) -> None:
        await self._interface.call_add_net_key(
            element_path, destination, subnet_index, net_index, update
        )

    async def add_app_key(
        self,
        element_path: str,
        destination: int,
        app_index: int,
        net_index: int,
        update: bool,
    ) -> None:
        await self._interface.call_add_app_key(
            element_path, destination, app_index, net_index, update
        )

    async def publish(self, element_path: str, model: int, data: bytes) -> None:
        await self._interface.call_publish(element_path, model, data)

    async def vendor_publish(
        self, element_path: str, vendor: int, model_id: int, data: bytes
    ):
        await self._interface.call_vendor_publish(element_path, vendor, model_id, data)

    async def update_sequence_number(self, seq_nr: int) -> int:
        updated_seq_nr = await self._interface.call_update_sequence_number(seq_nr)
        if updated_seq_nr != seq_nr:
            raise ValueError("Cannot update sequence number")
        return updated_seq_nr

    async def features(self) -> Mapping[str, bool]:
        return await self._interface.get_features()

    async def beacon(self) -> bool:
        return await self._interface.get_beacon()

    async def iv_update(self) -> bool:
        return await self._interface.get_iv_update()

    iv_update.signal = Signal()

    async def iv_index(self) -> int:
        return await self._interface.get_iv_index()

    iv_index.signal = Signal()

    async def iv_state(self) -> IvState:
        state = await self._interface.get_iv_state()
        return IvState(state)

    iv_state.signal = Signal()

    async def seconds_since_last_heard(self) -> int:
        return await self._interface.get_seconds_since_last_heard()

    async def address(self) -> int:
        return (await self._interface.get_addresses())[0]

    async def sequence_number(self) -> int:
        return await self._interface.get_sequence_number()


class ManagementInterface:
    def __init__(self, node_service):
        self._interface = node_service.get_interface("org.bluez.mesh.Management1")

    async def unprovisioned_scan(self, **kwargs) -> None:
        options = dict(Seconds=Variant("q", kwargs.get("seconds", 0)))
        await self._interface.call_unprovisioned_scan(options)

    async def unprovisioned_scan_cancel(self) -> None:
        await self._interface.call_unprovisioned_scan_cancel()

    async def add_node(self, uuid: UUID, **kwargs) -> None:
        options = dict(
            # TODO: populate with options when defined
        )
        await self._interface.call_add_node(uuid.bytes, options)

    async def create_subnet(self, net_index: int) -> None:
        await self._interface.call_create_subnet(net_index)

    async def import_subnet(self, net_index: int, net_key: NetworkKey) -> None:
        await self._interface.call_import_subnet(net_index, net_key.bytes)

    async def update_subnet(self, net_index: int) -> None:
        await self._interface.call_update_subnet(net_index)

    async def delete_subnet(self, net_index: int) -> None:
        await self._interface.call_delete_subnet(net_index)

    async def set_key_phase(self, net_index: int, phase: int) -> None:
        await self._interface.call_set_key_phase(net_index, phase)

    async def create_app_key(self, net_index: int, app_index: int) -> None:
        await self._interface.call_create_app_key(net_index, app_index)

    async def import_app_key(
        self, net_index: int, app_index: int, app_key: ApplicationKey
    ) -> None:
        await self._interface.call_import_app_key(net_index, app_index, app_key.bytes)

    async def update_app_key(self, app_index: int) -> None:
        await self._interface.call_update_app_key(app_index)

    async def complete_app_key_update(self, app_index: int) -> None:
        await self._interface.call_comlete_app_key_update(app_index)

    async def delete_app_key(self, app_index: int) -> None:
        await self._interface.call_delete_app_key(app_index)

    async def import_remote_node(
        self, primary: int, count: int, device_key: DeviceKey
    ) -> None:
        await self._interface.call_import_remote_node(primary, count, device_key.bytes)

    async def delete_remote_node(self, primary: int, count: int) -> None:
        await self._interface.call_delete_remote_node(primary, count)


class TCPServerService:
    NAME = "org.bluez.mesh"

    def __init__(self, port):
        self.PATH = f"/org/bluez/mesh/tcpserver_{port}"


class AccessControlListInterface:
    def __init__(self, tcpserver_service):
        self._interface = tcpserver_service.get_interface(
            "org.bluez.mesh.AccessControlList1"
        )

    async def grant_access(
        self, uuid: UUID, dev_key: DeviceKey, net_key: NetworkKey
    ) -> int:
        return await self._interface.call_grant_access(
            uuid.bytes, dev_key.bytes, net_key.bytes
        )

    async def revoke_access(self, token: int) -> None:
        await self._interface.call_revoke_access(token)


class TCPConnectionService:
    NAME = "org.bluez.mesh"

    def __init__(self, tcpserver_service: TCPServerService, identity: str):
        self.PATH = f"{tcpserver_service.PATH}/{identity}"


class ConnectionStatInterface:
    def __init__(self, tcpconnection_service):
        self._interface = tcpconnection_service.get_interface(
            "org.bluez.mesh.ConnectionStat1"
        )

    async def connected(self) -> bool:
        return self._interface.connected

    async def last_error(self) -> str:
        return self._interface.last_error

    async def transmitted_msg_count(self) -> int:
        return self._interface.transmitted_msg_count

    async def received_msg_count(self) -> int:
        return self._interface.received_msg_count

    async def last_transmitted_msg_timestamp(self) -> int:
        return self._interface.last_transmitted_msg_timestamp

    async def last_received_msg_timestamp(self) -> int:
        return self._interface.last_received_msg_timestamp



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
"""
This module provides a high-level API for BlueZ mesh stack.

.. _mesh-api.txt: https://git.kernel.org/pub/scm/bluetooth/bluez.git/tree/doc/mesh-api.txt
"""

import asyncio
import logging
import socket
import struct
from enum import Enum
from os import urandom
from functools import lru_cache, partial
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Union,
)
from uuid import UUID, uuid5

import construct
import dbus_next

from bluetooth_mesh.crypto import ApplicationKey, DeviceKey, NetworkKey
from bluetooth_mesh.interfaces import (
    AclInterface,
    ApplicationInterface,
    DBusInterface,
    DBusService,
    ElementInterface,
    ManagementInterface,
    MeshService,
    NetworkInterface,
    NodeInterface,
    ProvisionAgentInterface,
    ProvisionerInterface,
)
from bluetooth_mesh.messages import AccessMessage
from bluetooth_mesh.models import ConfigClient, ModelConfig
from bluetooth_mesh.tokenring import TokenRing
from bluetooth_mesh.utils import MeshError, ParsedMeshMessage

__all__ = [
    "Application",
    "Element",
]


class JoinComplete(NamedTuple):
    callback: Callable[[int], Awaitable]
    future: asyncio.Future


class Capabilities(Enum):
    BLINK = "blink"
    BEEP = "beep"
    VIBRATE = "vibrate"
    OUT_NUMERIC = "out-numeric"
    OUT_ALPHA = "out-alpha"
    PUSH = "push"
    TWIST = "twist"
    IN_NUMERIC = "in-numeric"
    IN_ALPHA = "in-alpha"
    STATIC_OOB = "static-oob"
    PUBLIC_OOB = "public-oob"


class OOBInfo(Enum):
    OTHER = "other"
    URI = "uri"
    MACHINE_CODE_2D = "machine-code-2d"
    BAR_CODE = "bar-code"
    NFC = "nfc"
    NUMBER = "number"
    STRING = "string"
    ON_BOX = "on-box"
    IN_BOX = "in-box"
    ON_PAPER = "on-paper"
    IN_MANUAL = "in-manual"
    ON_DEVICE = "on-device"


class CompositionDataMixin:
    """
    Provides propertied to company id, product id, version id and replay
    protection list size.

    These are used by :py:class:`bluetooth_mesh.interfaces.ApplicationInterface` to
    expose them via D-Bus.
    """

    COMPANY_ID = 0xFEE5
    PRODUCT_ID = 0x42
    VERSION_ID = 1
    CRPL = 0x7FFF

    @property
    def company_id(self) -> int:
        return self.COMPANY_ID

    @property
    def product_id(self) -> int:
        return self.PRODUCT_ID

    @property
    def version_id(self) -> int:
        return self.VERSION_ID

    @property
    def crpl(self) -> int:
        return self.CRPL


class PathMixin:
    """
    Provides `path` property under which
    py:class:`bluetooth_mesh.interfaces.ApplicationInterface` will be registered on
    D-Bus.
    """

    PATH = "/com/silvair/application"

    @property
    def path(self) -> str:
        return self.PATH


class MachineUUIDMixin(PathMixin):
    """
    Provides `uuid` property based on systemd's machine-id.

    """

    @staticmethod
    def get_namespace():
        with open("/etc/machine-id") as machine_id:
            return UUID(machine_id.read().strip())

    @property
    def uuid(self) -> UUID:
        namespace = self.get_namespace()
        return uuid5(namespace=namespace, name=self.path)


class TokenRingMixin(MachineUUIDMixin):
    """
    Provides a token ring handler based on application UUID.

    The handler is responsible for token ring persistence: the framework expects that tokens for nodes of
    a single application are stored & reused with subsequent launches.
    """

    TOKEN_RING = TokenRing

    @property
    @lru_cache(maxsize=1)
    def token_ring(self) -> TokenRing:
        return self.TOKEN_RING(uuid=self.uuid)


class NetworkKeyMixin:
    @property
    def primary_net_key(self) -> Tuple[int, NetworkKey]:
        """
        Index and key of the network that the application belongs to. Used when
        creating a new node, see :py:func:`Application.import_node`.
        """
        raise NotImplementedError("Getting primary network key should be overridden!")

    @property
    def subnet_keys(self) -> List[Tuple[int, NetworkKey]]:
        """
        Indexes and keys of the subnets.
        """
        raise NotImplementedError("Getting subnet network keys should be overridden!")


class AddressMixin:
    @property
    def address(self) -> int:
        addr = getattr(self, "__address", None)

        if addr is None:
            raise AttributeError("Application didn't provide an address")

        return addr

    @address.setter
    def address(self, value: int):
        if self.node_interface and getattr(self, "__address", None) != value:
            raise AttributeError("Can't set address once node is provisioned")

        setattr(self, "__address", value)


class IvIndexMixin:
    @property
    def iv_index(self) -> int:
        iv = getattr(self, "__iv_index", 0)

        if iv is None:
            raise AttributeError("Application didn't provide an IV Index")

        return iv

    @iv_index.setter
    def iv_index(self, value: int):
        if self.node_interface:
            raise AttributeError("Can't set IV Index once node is provisioned")

        setattr(self, "__iv_index", value)

    @property
    def iv_update(self) -> bool:
        iv = getattr(self, "__iv_update", False)

        if iv is None:
            raise AttributeError("Application didn't provide an IV Update")

        return iv

    @iv_update.setter
    def iv_update(self, value: bool):
        if self.node_interface:
            raise AttributeError("Can't set IV Update once node is provisioned")

        setattr(self, "__iv_update", value)


class ApplicationKeyMixin(NetworkKeyMixin):
    @property
    def primary_app_key(self) -> Tuple[int, int, ApplicationKey]:
        """
        Return first found application key that is bound to primary network key.
        """
        net_index, _ = self.primary_net_key

        for index, bound, key in self.app_keys:
            if bound == net_index:
                return index, bound, key

        raise IndexError("Primary application key not found")

    @property
    def app_keys(self) -> List[Tuple[int, int, ApplicationKey]]:
        """
        Indexes, bound network key indexes, and application keys.
        """
        raise NotImplementedError("Getting application keys should be overridden!")


class DeviceKeyMixin:
    @property
    def dev_key(self) -> DeviceKey:
        """
        Application's device_key. Used when creating a new node, see
        :py:func:`Application.import_node`.
        """
        return DeviceKey(urandom(16))


class DBusMixin:
    DBUS_SERVICE = None
    logger: logging.Logger

    def _name_owner_changed(self, name, old_owner, new_owner) -> None:
        if name != self.DBUS_SERVICE.NAME:
            return

        self.logger.error("Disconnected from %s (%s)", name, old_owner)
        self.dbus_disconnected(old_owner)

    async def dbus_connect(self) -> None:
        message_bus = dbus_next.aio.MessageBus(bus_type=dbus_next.BusType.SYSTEM)
        self.logger.debug("Connecting to D-Bus")
        self.bus = await message_bus.connect()

        introspection = await self.bus.introspect(DBusService.NAME, DBusService.PATH)
        dbus_service = self.bus.get_proxy_object(
            DBusService.NAME, DBusService.PATH, introspection
        )
        self.dbus_interface = DBusInterface(dbus_service)

        self.logger.info("Connecting to %s", self.DBUS_SERVICE.NAME)
        owner = await self.dbus_interface.get_service(self.DBUS_SERVICE)

        self.dbus_interface.on_name_owner_changed(self._name_owner_changed)
        await self.dbus_connected(owner)

    async def dbus_disconnect(self) -> None:
        self.dbus_interface.off_name_owner_changed(self._name_owner_changed)
        self.bus.disconnect()
        self.dbus_disconnected(None)

    async def dbus_connected(self, owner) -> None:
        pass

    def dbus_disconnected(self, owner) -> Any:
        self.loop.stop()

    async def __aenter__(self) -> "DBusMixin":
        await self.dbus_connect()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> Any:
        return await self.dbus_disconnect()


class ProvisioningMixin:
    CAPABILITIES = []
    OOBINFO = []
    URI = ""

    def private_key(self) -> bytes:
        """
        This method is called during provisioning if the Provisioner
        has requested Out-Of-Band ECC key exchange. The Private key is
        returned to the Daemon, and the Public Key is delivered to the
        remote Provisioner using a method that does not involve the
        Bluetooth Mesh system. The Private Key returned must be 32
        octets in size.
        """
        raise NotImplementedError("Getting private key should be overridden!")

    def public_key(self) -> bytes:
        """
        This method is called during provisioning if the local device is
        the Provisioner, and is requestng Out-Of-Band ECC key exchange.
        The Public key is returned to the Daemon that is the matched
        pair of the Private key of the remote device. The Public Key
        returned must be 64 octets in size.
        """
        raise NotImplementedError("Getting public key should be overridden!")

    def display_string(self, value: str):
        """
        This method is called when the Daemon has something important for the Agent to Display,
        but does not require any additional input locally.
        :param value: String
        :return:
        """
        raise NotImplementedError("Display functions should be overridden!")

    def display_numeric(self, type: str, number: int):
        """
        This method is called when the Daemon has something important
        for the Agent to Display, but does not require any additional
        input locally.
        :param type: String
        :param value: Integer
        :return:
        """
        raise NotImplementedError("Display functions should be overridden!")

    def prompt_static(self, type: str) -> bytes:
        """
        This method is called when the Daemon requires a 16 octet byte
        array, as an Out-of-Band authentication.
        :param type:
        :return:
        """
        raise NotImplementedError("Prompt functions should be overridden!")

    def prompt_numeric(self, type: str) -> int:
        """
        This method is called when the Daemon requests the user to
        enter a decimal value between 1-99999999.
        :param type:
        :return:
        """
        raise NotImplementedError("Prompt functions should be overridden!")

    def cancel(self):
        """
        This method gets called by the daemon to cancel any existing
        Agent Requests. When called, any pending user input should be
        canceled, and any display requests removed.
        :return:
        """
        raise NotImplementedError("Cancel functions should be overridden!")

    @property
    def capabilities(self) -> List[Capabilities]:
        """
        Return list of available capabilities.
        """
        return self.CAPABILITIES

    @capabilities.setter
    def capabilities(self, cap: List[Capabilities]):
        self.CAPABILITIES = cap

    @property
    def oob_info(self) -> List[OOBInfo]:
        """
        Indicates availability of OOB data.
        """
        return self.OOBINFO

    @oob_info.setter
    def oob_info(self, info: List[OOBInfo]):
        self.OOBINFO = info

    @property
    def uri(self) -> str:
        return self.URI


class ProvisionerMixin:
    def scan_result(self, rssi: int, data: bytes, options: dict):
        """
        The method is called from the bluetooth-meshd daemon when a
        unique UUID has been seen during UnprovisionedScan() for
        unprovsioned devices.
        :param rssi: signed, normalized measurement of the signal strength of the recieved unprovisioned beacon
        :param data:
        :param options:
        :return:
        """
        raise NotImplementedError("Provisioner functions should be overridden!")

    def request_prov_data(self, count: int) -> Tuple[int, int]:
        """
        This method is implemented by a Provisioner capable application
        and is called when the remote device has been fully
        authenticated and confirmed.

        :param count: consecutive unicast addresses the remote device is requesting
        :return:
            :param unet_index: Subnet index of the net_key
            :param uunicast: Primary Unicast address of the new node
        """
        raise NotImplementedError("Provisioner functions should be overridden!")

    def add_node_complete(self, uuid: bytes, unicast: int, count: int):
        """
        This method is called when the node provisioning initiated
        by an AddNode() method call successfully completed.

        :param uuid: 16 byte remote device UUID
        :param unicast: primary address that has been assigned to the new node, and the address of it's config server
        :param count: number of unicast addresses assigned to the new node
        :return:
        """
        raise NotImplementedError("Provisioner functions should be overridden!")

    def add_node_failed(self, uuid: bytes, reason: str):
        """
        This method is called when the node provisioning initiated by
        AddNode() has failed. Depending on how far Provisioning
        proceeded before failing, some cleanup of cached data may be
        required.

        :param uuid: 16 byte remote device UUID
        :param reason: reason for provisioning failure
        :return:
        """
        raise NotImplementedError("Provisioner functions should be overridden!")


class Application(
    CompositionDataMixin,
    TokenRingMixin,
    MachineUUIDMixin,
    PathMixin,
    ApplicationKeyMixin,
    DeviceKeyMixin,
    NetworkKeyMixin,
    DBusMixin,
    ProvisioningMixin,
    ProvisionerMixin,
    AddressMixin,
    IvIndexMixin,
):
    """
    Base class for mesh applications.

    """

    DBUS_SERVICE = MeshService

    ELEMENTS = {}  # type: Dict[int, Type[Element]]

    def __init__(self, loop: asyncio.AbstractEventLoop):
        super().__init__()

        self.loop = loop
        self.logger = logging.getLogger(type(self).__name__)

        self.application_interface = ApplicationInterface(self)
        self.provision_agent_interface = ProvisionAgentInterface(self)
        self.provisioner_interface = ProvisionerInterface(self)

        self.elements = {}  # type: Dict[int, Element]

        self.network_interface = None
        self.node_interface = None
        self.management_interface = None

        self._join_complete = None

    async def _get_acl_interface(self):
        mesh_introspection = await self.bus.introspect(
            MeshService.NAME, MeshService.PATH
        )
        tcp_server = [
            node.name
            for node in mesh_introspection.nodes
            if node.name.startswith("tcpserver_")
        ]
        if not tcp_server:
            self.logger.warning("TCP interface missing")
            raise NotImplementedError

        path = "%s/%s" % (self.DBUS_SERVICE.PATH, tcp_server[0])
        introspection = await self.bus.introspect(MeshService.NAME, path)
        acl_service = self.bus.get_proxy_object(MeshService.NAME, path, introspection)

        return AclInterface(acl_service)

    async def acl_grant(self, uuid, dev_key, net_key):
        server = await self._get_acl_interface()
        token = await server.grant_access(uuid.bytes, dev_key.bytes, net_key.bytes)
        self.token_ring.acl(uuid, token)

    async def acl_revoke(self, uuid):
        server = await self._get_acl_interface()
        await server.revoke_access(self.token_ring.acl(uuid))
        self.token_ring.drop_acl(uuid)

    async def dbus_connected(self, owner):
        introspection = await self.bus.introspect(
            self.DBUS_SERVICE.NAME, self.DBUS_SERVICE.PATH
        )
        service = self.bus.get_proxy_object(
            self.DBUS_SERVICE.NAME, self.DBUS_SERVICE.PATH, introspection
        )

        self.network_interface = NetworkInterface(service)
        self.node_interface = None
        self.management_interface = None

        self._register()

    def dbus_disconnected(self, owner):
        self._unregister()

        self.management_interface = None
        self.node_interface = None
        self.network_interface = None

        super().dbus_disconnected(owner)

    def _register(self):
        # pylint: disable=W0212
        self.logger.info("Registering application")

        self.bus.export(self.path, self.application_interface)
        self.bus.export(self.path, self.provision_agent_interface)
        self.bus.export(self.path, self.provisioner_interface)

        for index, element_class in self.ELEMENTS.items():
            element = element_class(self, index)
            self.elements[index] = element

            element_interface = ElementInterface(element)
            self.bus.export(element.path, element_interface)

    def _unregister(self):
        # pylint: disable=W0212
        self.logger.info("Unregistering application")

        for element in self.elements.values():
            self.bus.unexport(element.path)

        self.bus.unexport(self.path)

        self.elements = {}

    async def _join_callback(
        self, token, join_callback: Optional[Callable[[int], Awaitable[int]]] = None
    ):
        self.token_ring.token = token
        if join_callback:
            return await join_callback(token)

    async def connect(
        self,
        join_callback: Optional[Callable[[int], Awaitable[int]]] = None,
        **kwargs,
    ) -> Mapping[int, Dict[Tuple[int, int], Dict[str, Tuple[Any, int]]]]:
        """
        Connect to BlueZ. If a node doesn't exist yet, it gets created via
        Import() call, using self.dev_key, self.primary_net_key, self.address
        and self.iv_index.

        Returns current node configuration, see documentation for Attach()
        method in mesh-api.txt_.
        """
        try:
            configuration = await self.attach(self.token_ring.token, **kwargs)
        except (ValueError, dbus_next.errors.DBusError) as ex:
            self.logger.error("Attach failed: %s, trying to import node", ex)

            token = await self.import_node(
                join_callback=join_callback,
            )
            configuration = await self.attach(token, **kwargs)

        # after attaching, explicitly import own device key to enable
        # communication with local Config Server
        await self.management_interface.import_remote_node(
            self.address, len(self.ELEMENTS), self.dev_key
        )

        return configuration

    async def add_net_key(self, net_key_index: int, net_key: NetworkKey) -> Any:
        """
        Imports a network key into daemon's keyring.

        :param net_key_index: Index of the network key
        :param net_key: Value of the new network key
        """
        client = self.elements[0][ConfigClient]

        return await client.add_net_key(
            self.address,
            net_index=self.primary_net_key[0],
            net_key_index=net_key_index,
            net_key=net_key,
        )

    async def delete_net_key(self, net_key_index: int) -> Any:
        """
        Removes a network key from daemon's keyring.

        :param net_key_index: Index of the network key
        :param net_key: Value of the new network key
        """
        client = self.elements[0][ConfigClient]

        return await client.delete_net_key(
            self.address,
            net_index=self.primary_net_key[0],
            net_key_index=net_key_index,
        )

    async def add_app_key(
        self, net_key_index: int, app_key_index: int, app_key: ApplicationKey
    ) -> "AppKeyStatus":
        """
        Imports an application key into daemon's keyring.

        :param net_key_index: Index of the network key the new application key is bound to
        :param app_key_index: Index of the new application key
        :param app_key: Value of the new application key
        """
        client = self.elements[0][ConfigClient]

        return await client.add_app_key(
            self.address,
            net_index=self.primary_net_key[0],
            app_key_index=app_key_index,
            net_key_index=net_key_index,
            app_key=app_key,
        )

    async def delete_app_key(
        self, net_key_index: int, app_key_index: int
    ) -> "AppKeyStatus":
        """
        Removes an application key from daemon's keyring.

        :param net_key_index: Index of the network key the deleted application key is bound to
        :param app_key_index: Index of the deleted application key
        """
        client = self.elements[0][ConfigClient]

        return await client.delete_app_key(
            self.address,
            net_index=self.primary_net_key[0],
            app_key_index=app_key_index,
            net_key_index=net_key_index,
        )

    async def add_app_key(
        self, net_key_index: int, app_key_index: int, app_key: ApplicationKey
    ) -> "AppKeyStatus":
        """
        Imports an application key into daemon's keyring.

        :param net_key_index: Index of the network key the new application key is bound to
        :param app_key_index: Index of the new application key
        :param app_key: Value of the new application key
        """
        client = self.elements[0][ConfigClient]

        return await client.add_app_key(
            self.address,
            net_index=self.primary_net_key[0],
            app_key_index=app_key_index,
            net_key_index=net_key_index,
            app_key=app_key,
        )

    async def bind_app_key(
        self, app_key_index: int, model: "Model"
    ) -> "ModelBindStatus":
        """
        See :py:func:`Model.bind` instead.
        """
        client = self.elements[0][ConfigClient]

        return await client.bind_app_key(
            self.address,
            net_index=self.primary_net_key[0],
            element_address=self.address + model.element.index,
            app_key_index=app_key_index,
            model=type(model),
        )

    async def subscribe_model(
        self, subscription_address: int, model: "Model"
    ) -> "ModelSubscriptionStatus":
        """
        See :py:func:`Model.subscribe` instead.
        """
        client = self.elements[0][ConfigClient]

        return await client.add_subscription(
            self.address,
            net_index=self.primary_net_key[0],
            element_address=self.address + model.element.index,
            subscription_address=subscription_address,
            model=type(model),
        )

    async def unsubscribe_model(
        self, subscription_address: int, model: "Model"
    ) -> "ModelSubscriptionStatus":
        """
        See :py:func:`Model.unsubscribe` instead.
        """
        client = self.elements[0][ConfigClient]

        return await client.del_subscription(
            self.address,
            net_index=self.primary_net_key[0],
            element_address=self.address + model.element.index,
            subscription_address=subscription_address,
            model=type(model),
        )

    async def clear_subscriptions(self, model: "Model") -> "ModelSubscriptionStatus":
        """
        See :py:func:`Model.unsubscribe_all` instead.
        """
        client = self.elements[0][ConfigClient]

        return await client.clear_subscriptions(
            self.address,
            net_index=self.primary_net_key[0],
            element_address=self.address + model.element.index,
            model=type(model),
        )

    def get_model_instance(self, element: int, model: Type["Model"]) -> "Model":
        return self.elements[element][model]

    async def join(self):
        """
        Try to join a mesh network by broadcasting Unprovisioned Device Beacons,
        waiting for PB-ADV based provisioner.
        """
        self.logger.info("Join %s", self.uuid)

        self._join_complete = asyncio.Future()
        await self.network_interface.join("/", self.uuid)
        return await self._join_complete

    async def create_network(self):
        """
        Create a new mesh network.
        """
        self.logger.info("Create %s", self.uuid)

        self._join_complete = asyncio.Future()
        await self.network_interface.create_network("/", self.uuid)
        return await self._join_complete

    async def cancel(self):
        """
        Cancel outstanding :py:func:`join` request.
        """
        self.logger.info("Cancel")
        await self.network_interface.cancel()

    async def leave(self):
        """
        Remove the node.
        """
        self.logger.info("Leave")
        await self.network_interface.leave(self.token_ring.token)

    async def attach(self, token: int, *, socket_pair=False, socket_path: str = None):
        """
        Attach to existing node using a token.

        Returns current node configuration, see documentation for Attach()
        method in mesh-api.txt_.
        """
        self.logger.info(
            "Attach %x (socket_pair=%s, socket_path=%s)",
            token,
            socket_pair,
            socket_path,
        )

        if socket_pair and socket_path:
            raise AssertionError("Use either socket_pair or socket_path")

        if socket_pair:
            path, configuration, sock = await self.network_interface.attach_fd(
                "/", token
            )
            self._add_reader(sock)
        elif socket_path:
            path, configuration, sock = await self.network_interface.attach_unix(
                "/", token, socket_path
            )
            self._add_reader(sock)
        else:
            path, configuration = await self.network_interface.attach("/", token)

        introspection = await self.bus.introspect(MeshService.NAME, path)
        node_service = self.bus.get_proxy_object(MeshService.NAME, path, introspection)

        node_interface = NodeInterface(node_service)

        self.address = await node_interface.address()

        self.node_interface = node_interface
        self.management_interface = ManagementInterface(node_service)

        for element, models_configs in configuration.items():
            for model_id, model_config in models_configs.items():
                self.elements[element].update_model_configuration(
                    model_id, model_config
                )

        self.logger.info(
            "Attached to node %s, address: %04x, configuration: %s",
            path,
            self.address,
            configuration,
        )

        return configuration

    def _add_reader(self, sock: Any) -> Any:
        HEADER = struct.Struct(
            "<"
            "B"  # flags
            "H"  # source address
            "H"  # destination address
            "B"  # element index
            "H"  # appplication key index
            "H"  # net key index
            "B"  # ttl
            "16s"  # virtual label
            "Q"  # timestamp
        )

        def _read_message() -> Any:
            while True:
                try:
                    line, *_ = sock.recvmsg(1024)
                except BlockingIOError:
                    break

                header, data = line[: HEADER.size], line[HEADER.size :]
                (
                    flags,
                    source,
                    destination,
                    element,
                    app_index,
                    net_index,
                    ttl,
                    label,
                    _,
                ) = HEADER.unpack(header)

                dev_key = bool(flags & 0x01)

                if dev_key:
                    remote = bool(flags & 0x02)

                    self.elements[element].dev_key_message_received(
                        source, remote, net_index, data
                    )
                else:
                    self.elements[element].message_received(
                        source, app_index, destination, data
                    )

        self.loop.add_reader(sock, _read_message)

    async def import_node(
        self,
        join_callback: Optional[Callable[[int], Awaitable[int]]] = None,
        key_refresh: bool = False,
    ) -> int:
        """
        Create a self-provisioned node.
        """
        net_index, net_key = self.primary_net_key

        self.logger.warning(
            "Import node %s, address %04x, iv index %i%s",
            self.uuid.hex,
            self.address,
            self.iv_index,
            "(updating)" if self.iv_update else "",
        )

        flags = dict(
            IvUpdate=dbus_next.Variant("b", self.iv_update),
            KeyRefresh=dbus_next.Variant("b", key_refresh),
        )

        self._join_complete = JoinComplete(
            partial(self._join_callback, join_callback=join_callback), asyncio.Future()
        )
        await self.network_interface.import_node(
            "/", self.uuid, self.dev_key, net_key, net_index, flags, self.iv_index, self.address,
        )
        return await self._join_complete.future

    def join_complete(self, token: int):
        def join_complete_result(f: asyncio.Future):
            try:
                self._join_complete.future.set_result(token)
            except Exception as ex:
                self._join_complete.future.set_exception(ex)
                raise dbus_next.errors.DBusError(
                    "org.bluez.mesh.Application1", str(ex)
                ) from None

        join_task = self.loop.create_task(self._join_complete.callback(token))
        join_task.add_done_callback(join_complete_result)

    def join_failed(self, reason: str):
        self._join_complete.set_exception(MeshError(reason))


class LocationMixin:
    """
    Provides `location` property.
    """

    LOCATION = None  # type: int

    def __init__(self):
        assert self.LOCATION is not None

    @property
    def location(self) -> int:
        return self.LOCATION


class Element(LocationMixin):
    """
    Base class for elements.
    """

    MODELS = []  # type: List[Type["Model"]]

    def __init__(self, application: Application, index: int):
        super().__init__()

        self.logger = application.logger.getChild(
            "Element%d" % index
        )  # type: logging.Logger
        self.application = application
        self.index = index
        self.path = "%s/element%d" % (self.application.path, index)

        for opcode in sum((list(model.OPCODES) for model in self.MODELS), []):
            models = [model for model in self.MODELS if opcode in model.OPCODES]

            assert len(models) == 1, (
                "Element #%d declares models %r with overlapping opcode %r"
                % (index, models, opcode)
            )

        self._models = {
            model_class: model_class(self) for model_class in self.MODELS
        }  # type: Dict[Type["Model"], "Model"]

    def message_received(
        self, source: int, app_index: int, destination: Union[int, UUID], data: bytes
    ):
        """
        Called by :py:class:`bluetooth_mesh.interfaces.ElementInterface` when
        receiving a message encrypted with application key.

        The message is parsed using :py:class:`bluetooth_mesh.messages.AccessMessage`
        and (depending on the opcode) passed to relevant models'
        :func:`~bluetooth_mesh.models.Model.message_received`.

        """
        try:
            message = AccessMessage.parse(data)
        except construct.ConstructError as ex:
            self.logger.warning(
                "App message parse error [source %04x, app_index %d, destination %04x, data %s]: %s",
                source,
                app_index,
                destination,
                data.hex(),
                ex,
            )
            return

        for model in self._models.values():
            if message["opcode"] in model.OPCODES:
                model.message_received(source, app_index, destination, message)
                return

    def dev_key_message_received(
        self, source: int, remote: bool, net_index: int, data: bytes
    ):
        """
        Called by :py:class:`bluetooth_mesh.interfaces.ElementInterface` when
        receiving a message encrypted with device key.

        The message is parsed using :py:class:`bluetooth_mesh.messages.AccessMessage`
        and (depending on the opcode) passed to relevant models'
        :py:func:`bluetooth_mesh.models.Model.dev_key_message_received`.
        """

        try:
            message = AccessMessage.parse(data)
        except construct.ConstructError as ex:
            self.logger.warning(
                "Dev message parse error [source %04x, net_index %d, data %s]: %s",
                source,
                net_index,
                data.hex(),
                ex,
            )
            return

        for model in self._models.values():
            if message["opcode"] in model.OPCODES:
                model.dev_key_message_received(source, remote, net_index, message)
                return

    def update_model_configuration(
        self, model_id: Tuple[Optional[int], int], configuration: Mapping[str, Any]
    ):
        """
        Called by :py:class:`bluetooth_mesh.interfaces.ElementInterface` when model
        configuration is updated via daemon's internal Config Server model.

        Passes the configuration to relevant model's
        :py:func:`bluetooth_mesh.models.Model.update_configuration`.
        """
        for model in self._models.values():
            if model.MODEL_ID == model_id:
                model_config = ModelConfig(**configuration)
                model.update_configuration(model_config)
                return model_config

    def __getitem__(self, model_class: Type["Model"]) -> "Model":
        return self._models[model_class]

    @property
    def models(self) -> List[Tuple[int, bool, bool]]:
        """
        Used by :py:class:`bluetooth_mesh.interfaces.ApplicationInterface` to expose a
        list of supported SIG models via D-Bus.
        """
        return [
            (model.MODEL_ID[1], model.PUBLISH, model.SUBSCRIBE)
            for model in self._models.values()
            if model.MODEL_ID[0] is None
        ]

    @property
    def vendor_models(self) -> List[Tuple[Tuple[int, int], bool, bool]]:
        """
        Used by :py:class:`bluetooth_mesh.interfaces.ApplicationInterface` to expose a
        list of supported vendor models via D-Bus.
        """
        return [
            (model.MODEL_ID, model.PUBLISH, model.SUBSCRIBE)
            for model in self._models.values()
            if model.MODEL_ID[0] is not None
        ]

    def __repr__(self):
        def model_name(cls):
            if cls.MODEL_ID[0] is None:
                return "<%s %04x>" % (cls.__name__, cls.MODEL_ID[1])

            return "<%s %04x%04x>" % (cls.__name__, *cls.MODEL_ID)

        return "<%s: models=%s>" % (
            type(self).__name__,
            ", ".join(model_name(i) for i in self.MODELS),
        )



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    BitsInteger,
    BitStruct,
    Bytes,
    Const,
    Default,
    Embedded,
    Int8ul,
    Int16ul,
    PaddedString,
    Select,
    Struct,
    Switch,
    this,
)

from bluetooth_mesh.messages.util import (
    EnumAdapter,
    IpAddressAdapter,
    MacAddressAdapter,
    Opcode,
)


class GatewayConfigServerOpcode(IntEnum):
    SILVAIR_GATEWAY = 0xF03601


class GatewayConfigServerSubOpcode(IntEnum):
    GATEWAY_CONFIGURATION_GET = 0x00
    GATEWAY_CONFIGURATION_SET = 0x01
    GATEWAY_PACKETS_GET = 0x02
    GATEWAY_PACKETS_CLEAR = 0x03
    MTU_SIZE_SET = 0x04
    ETHERNET_MAC_ADDRESS_SET = 0x05
    SERVER_ADDRESS_AND_PORT_NUMBER_SET = 0x06
    RECONNECT_INTERVAL_SET = 0x07
    DNS_IP_ADDRESS_SET = 0x08
    IP_ADDRESS_SET = 0x09
    GATEWAY_IP_ADDRESS_SET = 0x0A
    NETMASK_SET = 0x0B
    GATEWAY_CONFIGURATION_STATUS = 0x0C
    GATEWAY_PACKETS_STATUS = 0x0D


class ConnState(IntEnum):
    ETHERNET_IDLE = 0x00
    ETHERNET_CONFIGURING = 0x01
    ETHERNET_INITIALIZING = 0x02
    ETHERNET_CONNECTING = 0x03
    ETHERNET_HANDSHAKE = 0x04
    ETHERNET_CONNECTED = 0x05


class LinkStatus(IntEnum):
    LINK_DOWN = 0x00
    LINK_UP = 0x01


class LastError(IntEnum):
    ERROR_NO_ERROR = 0x00
    ERROR_EPERM = 0x01
    ERROR_ENODEV = 0x02
    ERROR_EADDRINUSE = 0x03
    ERROR_EPROTO = 0x04
    ERROR_RFU1 = 0x05
    ERROR_RFU2 = 0x06
    ERROR_RFU3 = 0x07
    ERROR_RFU4 = 0x08
    ERROR_RFU5 = 0x09
    ERROR_RFU6 = 0x0A
    ERROR_RFU7 = 0x0B
    ERROR_RFU8 = 0x0C
    ERROR_RFU9 = 0x0D
    ERROR_RFU10 = 0x0E
    ERROR_UNKNOWN = 0x0F


class DhcpFlag(IntEnum):
    DHCP_DISABLED = 0x00
    DHCP_ENABLED_STATIC_DNS = 0x01
    DHCP_ENABLED_AUTO_DNS = 0x02


class StatusCode(IntEnum):
    STATUS_SUCCESS = 0x00
    STATUS_INVALID_ADDRESS = 0x01
    STATUS_INVALID_MODEL = 0x02
    STATUS_INVALID_APPKEY_IDX = 0x03
    STATUS_INVALID_NETKEY_IDX = 0x04
    STATUS_INSUFF_RESOURCES = 0x05
    STATUS_KEY_INDEX_ALREADY_STORED = 0x06
    STATUS_INVALID_PUBLISH_PARAMS = 0x07
    STATUS_NOT_SUBSCRIBE_MODEL = 0x08
    STATUS_STORAGE_FAILURE = 0x09
    STATUS_FEATURE_NOT_SUPPORTED = 0x0A
    STATUS_CANNOT_UPDATE = 0x0B
    STATUS_CANNOT_REMOVE = 0x0C
    STATUS_CANNOT_BIND = 0x0D
    STATUS_TEMP_UNABLE_TO_CHANGE_STATE = 0x0E
    STATUS_CANNOT_SET = 0x0F
    STATUS_UNSPECIFIED_ERROR = 0x10
    STATUS_INVALID_BINDING = 0x11


# fmt: off
# GATEWAY CONFIGURATION MSG
ConfigurationSetMtu = Struct(
    "mtu_size" / Int16ul,
)

ConfigurationSetMacAddr = Struct(
    "mac_address" / MacAddressAdapter(Bytes(6)),
)

ConfigurationSetServerAddrAndPortNr = Struct(
    "server_port_number" / Int16ul,
    "server_address_length" / Int8ul,
    "server_address" / PaddedString(this.server_address_length, "utf8"),
)

ConfigurationSetReconnectInterval = Struct(
    "reconnect_interval" / Int16ul,
)

ConfigurationSetDnsIpAddr = Struct(
    "dns_ip_address" / IpAddressAdapter(Bytes(4)),
)

ConfigurationSetIpAddr = Struct(
    "ip_address" / IpAddressAdapter(Bytes(4)),
)

ConfigurationSetGatewayIpAddr = Struct(
    "gateway_ip_address" / IpAddressAdapter(Bytes(4)),
)

ConfigurationSetNetmask = Struct(
    "netmask" / Int8ul,
)

ConfigurationSetWithoutOptionalAutoDhcpEnabled = Struct(
    "mtu_size" / Int16ul,
    "mac_address" / MacAddressAdapter(Bytes(6)),
    "server_port_number" / Int16ul,
    "reconnect_interval" / Int16ul,
    "server_address_length" / Int8ul,
    "server_address" / PaddedString(this.server_address_length, "utf8"),
)

ConfigurationSetWithOptionalDhcpEnabledWithStaticDns = Struct(
    Embedded(ConfigurationSetWithoutOptionalAutoDhcpEnabled),
    "dns_ip_address" / IpAddressAdapter(Bytes(4)),
)

ConfigurationSetWithOptionalDhcpDisabled = Struct(
    Embedded(ConfigurationSetWithOptionalDhcpEnabledWithStaticDns),
    "ip_address" / IpAddressAdapter(Bytes(4)),
    "gateway_ip_address" / IpAddressAdapter(Bytes(4)),
    "netmask" / Int8ul,
)

ConfigurationSet = Select(
    ConfigurationSetWithOptionalDhcpDisabled,
    ConfigurationSetWithOptionalDhcpEnabledWithStaticDns,
    ConfigurationSetWithoutOptionalAutoDhcpEnabled,
)

ConfigurationStatus = Struct(
    "chip_revision_id" / Int8ul,
    Embedded(ConfigurationSetWithOptionalDhcpDisabled),
    "flags" / EnumAdapter(Int8ul, DhcpFlag),
    "status_code" / EnumAdapter(Int8ul, StatusCode),
)

# GATEWAY PACKETS MSG
PacketsStatus = Struct(
    "total_eth_rx_errors" / Int16ul,
    "total_eth_tx_errors" / Int16ul,
    "bandwidth" / Int16ul,
    "connection_state" / BitStruct(
        "conn_state" / EnumAdapter(BitsInteger(3), ConnState),
        "link_status" / EnumAdapter(BitsInteger(1), LinkStatus),
        "last_error" / EnumAdapter(BitsInteger(4), LastError)
    )
)

GatewayConfigPayload = Struct(
    "subopcode" / EnumAdapter(Int8ul, GatewayConfigServerSubOpcode),
    "payload" / Default(Switch(
        this.subopcode,
        {
            GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET: ConfigurationSet,
            GatewayConfigServerSubOpcode.MTU_SIZE_SET: ConfigurationSetMtu,
            GatewayConfigServerSubOpcode.ETHERNET_MAC_ADDRESS_SET: ConfigurationSetMacAddr,
            GatewayConfigServerSubOpcode.SERVER_ADDRESS_AND_PORT_NUMBER_SET: ConfigurationSetServerAddrAndPortNr,
            GatewayConfigServerSubOpcode.RECONNECT_INTERVAL_SET: ConfigurationSetReconnectInterval,
            GatewayConfigServerSubOpcode.DNS_IP_ADDRESS_SET: ConfigurationSetDnsIpAddr,
            GatewayConfigServerSubOpcode.IP_ADDRESS_SET: ConfigurationSetIpAddr,
            GatewayConfigServerSubOpcode.GATEWAY_IP_ADDRESS_SET: ConfigurationSetGatewayIpAddr,
            GatewayConfigServerSubOpcode.NETMASK_SET: ConfigurationSetNetmask,
            GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS: ConfigurationStatus,
            GatewayConfigServerSubOpcode.GATEWAY_PACKETS_STATUS: PacketsStatus,
        },
    ), None)
)

GatewayConfigMessage = Struct(
    "opcode" / Const(GatewayConfigServerOpcode.SILVAIR_GATEWAY, Opcode(GatewayConfigServerOpcode)),
    "params" / GatewayConfigPayload
)
# fmt: on



from functools import partial

from bluetooth_mesh.crypto import ApplicationKey
from bluetooth_mesh.messages.config import ConfigOpcode


async def add_app_key(self,
                      destination: int,
                      net_index: int,
                      app_key_index: int,
                      net_key_index: int,
                      app_key: ApplicationKey):

    # create a Future that gets completed when APPKEY_STATUS is received
    status = self.expect_dev(
        destination,
        net_index=net_index,
        opcode=ConfigOpcode.APPKEY_STATUS,
        params=dict(
            app_key_index=app_key_index,
            net_key_index=net_key_index,
        )
    )

    # each request is the same, just keep retransmitting APPKEY_ADD
    request = partial(
        self.send_dev,
        destination,
        net_index=net_index,
        opcode=ConfigOpcode.APPKEY_ADD,
        params=dict(
            app_key_index=app_key_index,
            net_key_index=net_key_index,
            app_key=app_key.bytes,
        )
    )

    return await self.query(request, status, timeout=1.0)



import asyncio
import secrets
from contextlib import suppress

from docopt import docopt

from bluetooth_mesh.application import Application, Element
from bluetooth_mesh.crypto import ApplicationKey, DeviceKey, NetworkKey
from bluetooth_mesh.messages.config import GATTNamespaceDescriptor, StatusCode
from bluetooth_mesh.models import ConfigClient, HealthClient


class MainElement(Element):
    LOCATION = GATTNamespaceDescriptor.MAIN
    MODELS = [
        ConfigClient,
        HealthClient,
    ]


class SampleApplication(Application):
    COMPANY_ID = 0x0136  # Silvair
    PRODUCT_ID = 1
    VERSION_ID = 1
    ELEMENTS = {
        0: MainElement,
    }
    CRPL = 32768
    PATH = "/com/silvair/sample"

    @property
    def dev_key(self):
        return DeviceKey(secrets.token_bytes(16))

    @property
    def net_key(self):
        return 0, NetworkKey(secrets.token_bytes(16))

    @property
    def app_keys(self):
        return {0: ApplicationKey(secrets.token_bytes(16))}

    async def configure(self):
        client = self.elements[0][ConfigClient]

        status = await client.add_app_key(
            self.addr, net_index=0,
            app_key_index=0,
            net_key_index=0,
            app_key=self.app_keys[0]
        )

        assert status == StatusCode.SUCCESS, \
            'Cannot add application key: %s' % status

        status = await client.bind_app_key(
            self.addr, net_index=0,
            element_address=self.addr,
            app_key_index=0,
            model=HealthClient
        )

        assert status == StatusCode.SUCCESS, \
            'Cannot bind application key: %s' % status

    async def run(self, addr):
        async with self:
            await self.connect(addr=addr, iv_index=5)
            await self.configure()

            client = self.elements[0][HealthClient]

            for node in [0x0001, 0x0002, 0x0003]:
                await client.attention(node, app_index=0, attention=3)


def main():
    doc = """
    Sample Application

    Usage:
        sample --address <address>
        meshcli -h | --help | --version

    Options:
        -a --address <address>          Local node unicast address
        -h --help                       Show this help message and exit
        --version                       Show version and exit
    """

    arguments = docopt(doc, version='1.0')

    addr = arguments['--address']
    addr = int(addr, 16 if addr.startswith('0x') else 10)

    loop = asyncio.get_event_loop()
    app = SampleApplication(loop)

    with suppress(KeyboardInterrupt):
        loop.run_until_complete(app.run(addr))


if __name__ == '__main__':
    main()



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0223

import math
from ipaddress import IPv4Address

from construct import (
    Adapter,
    Bit,
    BitStruct,
    Bitwise,
    Computed,
    Construct,
    Container,
    Embedded,
    Enum,
    ExprValidator,
    Int8ub,
    Int16ub,
    Int24ub,
    Rebuild,
    Restreamed,
    Select,
    SizeofError,
    Struct,
    Switch,
    ValidationError,
    obj_,
    stream_read,
    stream_write,
    this,
)


def identity(x):
    return x


def reverse(sequence):
    return sequence[::-1]


def Reversed(subcon):
    return Restreamed(
        subcon, reverse, subcon.sizeof(), reverse, subcon.sizeof(), identity
    )


class BitListAdapter(Adapter):
    def __init__(self, subcon, reversed):
        super().__init__(subcon)
        self.reversed = reversed

    def _decode(self, obj, content, path):
        bits = set()

        for bit, feature in enumerate(reversed(obj) if self.reversed else obj):
            if not feature:
                continue

            bits.add(bit)

        return bits

    def _encode(self, obj, content, path):
        bits = []

        for bit in range(self.sizeof() * 8):
            bits.append(bit in obj)

        return list(reversed(bits)) if self.reversed else bits


def BitList(size, *, reversed=False):
    return BitListAdapter(Bitwise(Bit[size * 8]), reversed)


def EnumAdapter(subcon, enum):
    class _Enum(Enum):
        ENUM = enum

    class _EnumAdapter(Adapter):
        def _decode(self, obj, context, path):
            if obj not in enum._value2member_map_:
                raise ValidationError("object failed validation: %s" % (obj,))
            return enum(obj)

        def _encode(self, obj, context, path):
            if obj == enum:
                return obj.value

            try:
                return enum[obj] if isinstance(obj, str) else enum(obj)
            except ValueError:
                raise ValidationError("object failed validation: %s" % (obj,))

    _EnumAdapter.__construct_doc__ = _Enum(subcon, enum)

    return _EnumAdapter(subcon)


def LogAdapter(subcon, *, max_value=None, infinity=False):
    class _LogAdapter(Adapter):
        MAX_TYPE_VALUE = int(math.pow(2, subcon.length * 8) - 1)

        def _decode(self, obj, context, path):
            if obj == 0:
                return 0

            if obj == self.MAX_TYPE_VALUE:
                if infinity:
                    return float("inf")

            if max_value is not None and obj > max_value:
                raise ValidationError(
                    "max value exceeded, expecting at most %d: %s" % (max_value, obj)
                )

            return int(math.pow(2, obj - 1))

        def _encode(self, obj, context, path):
            if obj == 0:
                return 0

            if obj == float("inf"):
                if infinity:
                    return self.MAX_TYPE_VALUE

                raise ValidationError("infinity is not allowed: %s" % obj)

            value = math.log(obj, 2) + 1

            if max_value is not None and value > max_value:
                raise ValidationError(
                    "max value exceeded, expecting at most %d: %s" % (max_value, obj)
                )

            return int(value)

    return _LogAdapter(subcon)


def RangeValidator(subcon, *, min_value=None, max_value=None):
    def validate_range(obj, ctx):
        if min_value is not None and obj < min_value:
            return False

        if max_value is not None and obj > max_value:
            return False

        return True

    return ExprValidator(subcon, validate_range)


def EmbeddedBitStruct(name, *args, reversed=False):
    """
    Emulates BitStruct embedding:
        - for parsing, adds Computed accessor fields to the parent construct,
        - for building, Rebuild the bit struct using keys passed to the parent

    NOTE: This is a hack. Do not use unless you absolutely have to.
    """
    bit_struct = BitStruct(*args)

    if reversed:
        bit_struct = Reversed(bit_struct)

    bit_struct.__construct_doc__ = Embedded(Struct(*args))

    return (name / Rebuild(bit_struct, dict),) + tuple(
        i.name / Computed(this[name][i.name]) for i in args if i.name is not None
    )


class Opcode(Construct):
    def __init__(self, type=int):
        super().__init__()
        self.type = type

    def _parse(self, stream, context, path):
        try:
            opcode = stream_read(stream, 1)[0]

            if opcode == 0x7F:
                raise ValidationError

            len = opcode >> 7

            # 1 byte opcode
            if not len:
                return self.type(opcode)

            len = opcode >> 6
            opcode = opcode << 8 | stream_read(stream, 1)[0]

            # 2 byte opcode
            if len == 2:
                return self.type(opcode)

            if len == 3:
                opcode = opcode << 8 | stream_read(stream, 1)[0]
                return self.type(opcode)

            raise ValidationError
        except ValueError:
            raise ValidationError

    def _build(self, obj, stream, context, path):
        if obj > 0xFFFF:
            stream_write(stream, obj.to_bytes(3, byteorder="big"))
        elif obj > 0xFF:
            stream_write(stream, obj.to_bytes(2, byteorder="big"))
        else:
            stream_write(stream, obj.to_bytes(1, byteorder="big"))

        return self.type(obj)

    def _sizeof(self, context, path):
        raise SizeofError


class DefaultCountValidator(Adapter):
    def __init__(self, subcon, rounding=None, resolution=1.0):
        super().__init__(subcon)
        self.rounding = rounding
        self.resolution = resolution

    def _decode(self, obj, content, path):
        if obj == (256 ** self.subcon.length) - 1:
            return None
        else:
            return (
                round(obj * self.resolution, self.rounding)
                if self.rounding
                else obj * self.resolution
            )

    def _encode(self, obj, content, path):
        if obj is None:
            return (256 ** self.subcon.length) - 1
        else:
            return round(obj / self.resolution)


# TODO: is there a better way to calculate size of a nested struct?
def SwitchLength(type, switch, mapping, path, default=0):
    def rebuild(type, v):
        return Rebuild(type, lambda x: len(v.build(path(x))))

    return Switch(
        switch,
        {k: rebuild(type, v) for k, v in mapping.items()},
        default=Rebuild(type, lambda x: default),
    )


def RebuildLength(type, subcons, path):
    return Rebuild(type, lambda x: len(subcons.build(path(x))))


class MacAddressAdapter(Adapter):
    def _decode(self, obj, context, path):
        return ":".join(map("{:02x}".format, obj))

    def _encode(self, obj, context, path):
        return bytes(int(i, 16) for i in obj.split(":"))


class IpAddressAdapter(Adapter):
    def _decode(self, obj, context, path):
        return IPv4Address(obj)

    def _encode(self, obj, context, path):
        return bytes(int(i) for i in obj.split("."))


class DictAdapter(Adapter):
    def __init__(self, subcon, key, value):
        super().__init__(subcon)
        self.key = key
        self.value = value

    def _decode(self, obj, context, path):
        if isinstance(self.value, (list, tuple)):
            return {
                self.key(i): {v.__getfield__(): v(i) for v in self.value} for i in obj
            }

        return {self.key(i): self.value(i) for i in obj}

    def _encode(self, obj, context, path):
        key_name = self.key.__getfield__()

        if isinstance(self.value, (list, tuple)):
            for k, v in obj.items():
                yield {key_name: k, **v}
        else:
            value_name = self.value.__getfield__()
            for k, v in obj.items():
                yield {key_name: k, value_name: v}


class OpcodeMessage(Construct):
    OPCODE = Opcode()

    def __init__(self, opcodes):
        super().__init__()
        self.opcodes = {k: v.compile() for k, v in opcodes.items()}

    def _parse(self, stream, context, path):
        opcode = self.OPCODE._parse(stream, context, path)
        params = self.opcodes[opcode]._parse(stream, context, path)

        return Container(opcode=opcode, params=params)

    def _build(self, obj, stream, context, path):
        opcode = obj["opcode"]

        self.OPCODE._build(opcode, stream, context, path)
        self.opcodes[opcode]._build(obj["params"], stream, context, path)

        return obj

    def _sizeof(self, context, path):
        raise SizeofError



from typing import Any, Mapping, Union
from uuid import UUID

from bluetooth_mesh.application import Element
from bluetooth_mesh.messages.generic.onoff import GenericOnOffOpcode
from bluetooth_mesh.models import Model


class MyModel(Model):
    MODEL_ID = (None, 0x1002)
    OPCODES = [
        GenericOnOffOpcode.ONOFF_SET,
    ]

    def _app_onoff_set(source: int,
                       app_index: int,
                       destination: Union[int, UUID],
                       message: Mapping[str, Any]):
        pass

    def _dev_onoff_set(source: int,
                       net_index: int,
                       message: Mapping[str, Any]):
        pass

    def __init__(self, element: Element):
        super().__init__(element)

        self.app_message_callbacks[GenericOnOffOpcode.ONOFF_SET] \
            .add(self._app_onoff_set)

        self.dev_message_callbacks[GenericOnOffOpcode.ONOFF_SET] \
            .add(self._dev_onoff_set)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0621

import asyncio
from asyncio import Future
from functools import partial
from unittest.mock import MagicMock, call

import asynctest
import pytest

from bluetooth_mesh import Model
from bluetooth_mesh.interfaces import NodeInterface
from bluetooth_mesh.messages.generic.onoff import GenericOnOffOpcode
from bluetooth_mesh.test.fixtures import *  # pylint: disable=W0614, W0401


@pytest.fixture
def node_interface():
    return asynctest.MagicMock(NodeInterface)


@pytest.fixture
def element_mock(element_path, node_interface):
    element_mock = MagicMock()
    element_mock.application.node_interface = node_interface
    element_mock.path = element_path
    return element_mock


@pytest.fixture()
def model(element_mock):
    class MockModel(Model):
        MODEL_ID = (None, 0x1001)
        OPCODES = {GenericOnOffOpcode.ONOFF_STATUS}

    return MockModel(element_mock)


@pytest.mark.asyncio
async def test_single_expect(model, status_parsed, source, app_index):
    status = model.expect_app(source, app_index, False, status_parsed["opcode"], {})
    model.message_received(source, app_index, False, status_parsed)
    assert status_parsed == await status


@pytest.mark.asyncio
async def test_wildcard_expect(model, status_parsed, source, app_index):
    status = model.expect_app(
        source, app_index, False, status_parsed["opcode"], dict(present_onoff=...)
    )
    model.message_received(source, app_index, False, status_parsed)
    assert status_parsed == await status


@pytest.mark.asyncio
async def test_multiple_expect(model, status_parsed, source, app_index):
    status1 = model.expect_app(source, app_index, False, status_parsed["opcode"], {})
    status2 = model.expect_app(source, app_index, False, status_parsed["opcode"], {})
    model.message_received(source, app_index, False, status_parsed)
    assert status_parsed == await status1
    assert status_parsed == await status2


def test_app_single_repeatable_callback(model, status_parsed, source, app_index):
    listener_mock = MagicMock(return_value=False)
    model.app_message_callbacks[status_parsed["opcode"]].add(listener_mock)
    model.message_received(source, app_index, False, status_parsed)
    listener_mock.assert_called_once_with(source, app_index, False, status_parsed)

    listener_mock.reset_mock()
    model.message_received(source, app_index, False, status_parsed)
    listener_mock.assert_called_once_with(source, app_index, False, status_parsed)


def test_single_oneshot_callback(model, status_parsed, source, app_index):
    listener_mock = MagicMock(return_value=True)
    model.app_message_callbacks[status_parsed["opcode"]].add(listener_mock)
    model.message_received(source, app_index, False, status_parsed)
    listener_mock.assert_called_once_with(source, app_index, False, status_parsed)

    listener_mock.reset_mock()
    model.message_received(source, app_index, False, status_parsed)
    listener_mock.assert_not_called()


def test_multiple_callback(model, status_parsed, source, app_index):
    listener_mock_repeatable = MagicMock(return_value=False)
    listener_mock_oneshot = MagicMock(return_value=True)
    model.app_message_callbacks[status_parsed["opcode"]].add(listener_mock_repeatable)
    model.app_message_callbacks[status_parsed["opcode"]].add(listener_mock_oneshot)
    model.message_received(source, app_index, False, status_parsed)
    listener_mock_repeatable.assert_called_once_with(
        source, app_index, False, status_parsed
    )
    listener_mock_oneshot.assert_called_once_with(
        source, app_index, False, status_parsed
    )

    listener_mock_repeatable.reset_mock()
    listener_mock_oneshot.reset_mock()
    model.message_received(source, app_index, False, status_parsed)
    listener_mock_repeatable.assert_called_once_with(
        source, app_index, False, status_parsed
    )
    listener_mock_oneshot.assert_not_called()


@pytest.mark.asyncio
async def test_dev_single_expect(model, status_parsed, source, net_index):
    status = model.expect_dev(source, net_index, status_parsed["opcode"], {})
    model.dev_key_message_received(source, True, net_index, status_parsed)
    assert status_parsed == await status


@pytest.mark.asyncio
async def test_dev_wildcard_expect(model, status_parsed, source, net_index):
    status = model.expect_dev(
        source, net_index, status_parsed["opcode"], dict(present_onoff=...)
    )
    model.dev_key_message_received(source, True, net_index, status_parsed)
    assert status_parsed == await status


@pytest.mark.asyncio
async def test_dev_multiple_expect(model, status_parsed, source, net_index):
    status1 = model.expect_dev(source, net_index, status_parsed["opcode"], {})
    status2 = model.expect_dev(source, net_index, status_parsed["opcode"], {})
    model.dev_key_message_received(source, True, net_index, status_parsed)
    assert status_parsed == await status1
    assert status_parsed == await status2


def test_dev_single_repeatable_callback(model, status_parsed, source, net_index):
    listener_mock = MagicMock(return_value=False)
    model.dev_message_callbacks[status_parsed["opcode"]].add(listener_mock)
    model.dev_key_message_received(source, True, net_index, status_parsed)
    listener_mock.assert_called_once_with(source, net_index, status_parsed)

    listener_mock.reset_mock()
    model.dev_key_message_received(source, True, net_index, status_parsed)
    listener_mock.assert_called_once_with(source, net_index, status_parsed)


def test_dev_single_oneshot_callback(model, status_parsed, source, net_index):
    listener_mock = MagicMock(return_value=True)
    model.dev_message_callbacks[status_parsed["opcode"]].add(listener_mock)
    model.dev_key_message_received(source, True, net_index, status_parsed)
    listener_mock.assert_called_once_with(source, net_index, status_parsed)

    listener_mock.reset_mock()
    model.dev_key_message_received(source, True, net_index, status_parsed)
    listener_mock.assert_not_called()


def test_dev_multiple_callback(model, status_parsed, source, net_index):
    listener_mock_repeatable = MagicMock(return_value=False)
    listener_mock_oneshot = MagicMock(return_value=True)
    model.dev_message_callbacks[status_parsed["opcode"]].add(listener_mock_repeatable)
    model.dev_message_callbacks[status_parsed["opcode"]].add(listener_mock_oneshot)
    model.dev_key_message_received(source, True, net_index, status_parsed)
    listener_mock_repeatable.assert_called_once_with(source, net_index, status_parsed)
    listener_mock_oneshot.assert_called_once_with(source, net_index, status_parsed)

    listener_mock_oneshot.reset_mock()
    listener_mock_repeatable.reset_mock()
    model.dev_key_message_received(source, True, net_index, status_parsed)
    listener_mock_oneshot.assert_not_called()
    listener_mock_repeatable.assert_called_once_with(source, net_index, status_parsed)


@pytest.mark.asyncio
async def test_send_app(
    model, status_parsed, destination, app_index, element_path, node_interface
):
    await model.send_app(
        destination, app_index, status_parsed["opcode"], status_parsed["params"]
    )
    node_interface.send.assert_called_once_with(
        element_path, destination, app_index, b"\x82\x04\x00"
    )


@pytest.mark.asyncio
async def test_send_dev(
    model, status_parsed, destination, net_index, element_path, node_interface
):
    await model.send_dev(
        destination, net_index, status_parsed["opcode"], status_parsed["params"]
    )
    node_interface.dev_key_send.assert_called_once_with(
        element_path, destination, True, net_index, b"\x82\x04\x00"
    )


@pytest.mark.asyncio
async def test_repeat(
    model, status_parsed, destination, app_index, element_path, node_interface
):
    request = partial(
        model.send_app,
        destination,
        app_index,
        status_parsed["opcode"],
        status_parsed["params"],
    )
    await model.repeat(request, send_interval=0.0001)
    node_interface.send.assert_has_calls(
        [
            asynctest.call(element_path, destination, app_index, b"\x82\x04\x00")
            for _ in range(6)
        ]
    )
    assert node_interface.send.call_count == 6


@pytest.mark.asyncio
async def test_query(
    model, status_parsed, source, destination, app_index, element_path, node_interface
):
    request_future = Future()

    async def request():
        await model.send_app(
            destination, app_index, status_parsed["opcode"], status_parsed["params"]
        )
        request_future.set_result(None)

    status = model.expect_app(source, app_index, False, status_parsed["opcode"], {})
    query_status = asyncio.ensure_future(model.query(request, status))
    await request_future
    model.message_received(source, app_index, False, status_parsed)
    await query_status
    node_interface.send.assert_called_once_with(
        element_path, destination, app_index, b"\x82\x04\x00"
    )


@pytest.mark.asyncio
async def test_bulk_query(
    model, status_parsed, destination, app_index, element_path, node_interface
):
    async def request(future, dest):
        await model.send_app(
            dest, app_index, status_parsed["opcode"], status_parsed["params"]
        )
        future.set_result(None)

    destinations = [destination + i for i in range(10)]
    futures = {addr: Future() for addr in destinations}
    requests = {addr: partial(request, futures[addr], addr) for addr in destinations}
    statuses = {
        addr: model.expect_app(addr, app_index, False, status_parsed["opcode"], {})
        for addr in destinations
    }

    query_status = asyncio.ensure_future(
        model.bulk_query(requests, statuses, send_interval=0.01)
    )
    await asyncio.wait(futures.values())

    for addr in destinations:
        model.message_received(addr, app_index, False, status_parsed)

    await query_status
    node_interface.send.assert_has_calls(
        [call(element_path, addr, app_index, b"\x82\x04\x00") for addr in destinations]
    )
    assert node_interface.send.call_count == 10


@pytest.mark.asyncio
async def test_bulk_query_timeout(
    model, status_parsed, destination, app_index, element_path, node_interface
):
    async def request(dest):
        await model.send_app(
            dest, app_index, status_parsed["opcode"], status_parsed["params"]
        )

    async def callback(address, result, done, total):
        if address == 10:
            assert len(done) == 1
            assert result == {
                "opcode": GenericOnOffOpcode.ONOFF_STATUS,
                "params": {"present_onoff": 0},
            }
        elif address == 20:
            assert len(done) == 2
            assert isinstance(result, TimeoutError)

    requests = {
        10: partial(request, 10),
        20: partial(request, 20),
    }
    statuses = {
        10: model.expect_app(10, app_index, False, status_parsed["opcode"], {}),
        20: model.expect_app(20, app_index, False, status_parsed["opcode"], {}),
    }

    query_status = asyncio.ensure_future(
        model.bulk_query(
            requests,
            statuses,
            send_interval=0.001,
            progress_callback=callback,
            timeout=0.1,
        )
    )

    model.message_received(10, app_index, False, status_parsed)

    await query_status



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import uuid

from pytest import fixture, raises, skip

from bluetooth_mesh.crypto import ApplicationKey, DeviceKey, NetworkKey
from bluetooth_mesh.mesh import (
    AccessMessage,
    ControlMessage,
    NetworkMessage,
    Nonce,
    ProxyConfigMessage,
    SecureNetworkBeacon,
    SegmentAckMessage,
    SolicitationMessage,
    UnprovisionedDeviceBeacon,
)


@fixture
def app_key():
    return ApplicationKey(bytes.fromhex("63964771734fbd76e3b40519d1d94a48"))


@fixture
def net_key():
    return NetworkKey(bytes.fromhex("7dd7364cd842ad18c17c2b820c84c3d6"))


@fixture
def dev_key():
    return DeviceKey(bytes.fromhex("9d6dd0e96eb25dc19a40ed9914f8f03f"))


@fixture
def health_current_status_message():
    payload = bytes.fromhex("0400000000")
    return AccessMessage(src=0x1201, dst=0xFFFF, ttl=0x03, payload=payload)


@fixture
def config_appkey_status_message():
    payload = bytes.fromhex("800300563412")
    return AccessMessage(src=0x1201, dst=0x0003, ttl=0x0B, payload=payload)


@fixture
def config_appkey_add_message():
    payload = bytes.fromhex("0056341263964771734fbd76e3b40519d1d94a48")
    return AccessMessage(src=0x0003, dst=0x1201, ttl=0x04, payload=payload)


@fixture
def control_appkey_add_ack_message():
    return SegmentAckMessage(
        src=0x2345, dst=0x0003, ttl=0x0B, seq_zero=0x09AB, ack_segments=[1], obo=True
    )


@fixture
def control_friend_offer_message():
    parameters = bytes.fromhex("320308ba072f")
    return ControlMessage(
        src=0x2345, dst=0x1201, ttl=0x00, opcode=0x04, payload=parameters
    )


@fixture
def proxy_use_whitelist_message():
    parameters = bytes.fromhex("00")
    return ProxyConfigMessage(src=0x0001, opcode=0x00, payload=parameters)


@fixture
def app_sar_message():
    payload = bytes.fromhex("510c00000000020b0c1f00efcdab071b1c")
    return AccessMessage(src=1234, dst=4321, ttl=1, payload=payload)


@fixture
def proxy_solicitation_message():
    return SolicitationMessage(src=0x0031, dst=0x0100)


def test_network_beacon_received(net_key):
    beacon, auth = SecureNetworkBeacon.unpack(
        bytes.fromhex("003ecaff672f673370123456788ea261582f364f6f")
    )

    assert beacon.verify(auth, net_key)

    assert beacon.network_id == net_key.network_id
    assert not beacon.key_refresh
    assert not beacon.iv_update
    assert beacon.iv_index == 0x12345678


def test_network_beacon_received_iv_update(net_key):
    beacon, auth = SecureNetworkBeacon.unpack(
        bytes.fromhex("023ecaff672f67337012345679c2af80ad072a135c")
    )

    assert beacon.verify(auth, net_key)

    assert beacon.network_id == net_key.network_id
    assert not beacon.key_refresh
    assert beacon.iv_update
    assert beacon.iv_index == 0x12345679


def test_network_beacon_created(net_key):
    beacon = SecureNetworkBeacon(
        False, False, 0x12345678, network_id=net_key.network_id
    )

    assert beacon.pack(net_key) == (
        bytes.fromhex("003ecaff672f67337012345678"),
        bytes.fromhex("8ea261582f364f6f"),
    )


def test_unprovisioned_beacon_received():
    beacon = UnprovisionedDeviceBeacon.unpack(
        bytes.fromhex("25bdf2eb03cc4383a65add3e8007fb554243")
    )

    assert beacon.uuid == uuid.UUID("25bdf2eb-03cc-4383-a65a-dd3e8007fb55")
    assert beacon.oob == 0x4243


def test_unprovisioned_beacon_received_uri_hash():
    beacon = UnprovisionedDeviceBeacon.unpack(
        bytes.fromhex("25bdf2eb03cc4383a65add3e8007fb55424301020304")
    )

    assert beacon.uuid == uuid.UUID("25bdf2eb-03cc-4383-a65a-dd3e8007fb55")
    assert beacon.oob == 0x4243
    assert beacon.uri_hash == bytes.fromhex("01020304")


def test_unprovisioned_beacon_received_uri_hash_too_short():
    with raises(ValueError, match="expected 4 bytes"):
        UnprovisionedDeviceBeacon.unpack(
            bytes.fromhex("25bdf2eb03cc4383a65add3e8007fb554243010203")
        )


def test_unprovisioned_beacon_created():
    beacon = UnprovisionedDeviceBeacon(
        uuid.UUID("25bdf2eb-03cc-4383-a65a-dd3e8007fb55"), 0x4243
    )

    assert beacon.pack() == bytes.fromhex("25bdf2eb03cc4383a65add3e8007fb554243")


def test_unprovisioned_beacon_created_uri_hash():
    beacon = UnprovisionedDeviceBeacon(
        uuid.UUID("25bdf2eb-03cc-4383-a65a-dd3e8007fb55"),
        0x4243,
        bytes.fromhex("04030201"),
    )

    assert beacon.pack() == bytes.fromhex(
        "25bdf2eb03cc4383a65add3e8007fb55424304030201"
    )


def test_unprovisioned_beacon_created_uri_hash_too_short():
    with raises(ValueError, match="expected 4 bytes"):
        beacon = UnprovisionedDeviceBeacon(
            uuid.UUID("25bdf2eb-03cc-4383-a65a-dd3e8007fb55"),
            0x4243,
            bytes.fromhex("040302"),
        )


def test_network_nonce(config_appkey_status_message):
    assert Nonce(
        config_appkey_status_message.src,
        config_appkey_status_message.dst,
        config_appkey_status_message.ttl,
        False,
    ).network(seq=0x000006, iv_index=0x12345678) == bytes.fromhex(
        "000b0000061201000012345678"
    )


def test_device_nonce(config_appkey_status_message):
    assert Nonce(
        config_appkey_status_message.src,
        config_appkey_status_message.dst,
        config_appkey_status_message.ttl,
        False,
    ).device(seq=0x000006, iv_index=0x12345678) == bytes.fromhex(
        "02000000061201000312345678"
    )


def test_application_nonce(health_current_status_message):
    assert Nonce(
        health_current_status_message.src,
        health_current_status_message.dst,
        health_current_status_message.ttl,
        False,
    ).application(seq=0x000007, iv_index=0x12345678) == bytes.fromhex(
        "01000000071201ffff12345678"
    )


def test_device_pdu(config_appkey_status_message, dev_key):
    (segment,) = config_appkey_status_message.segments(
        dev_key, seq=0x000006, iv_index=0x12345678
    )

    assert segment == bytes.fromhex("0089511bf1d1a81c11dcef")


def test_application_pdu(health_current_status_message, app_key):
    (segment,) = health_current_status_message.segments(
        app_key, seq=0x000007, iv_index=0x12345678
    )

    assert segment == bytes.fromhex("665a8bde6d9106ea078a")


def test_application_pdu_segmented(config_appkey_add_message, dev_key):
    segments = list(
        config_appkey_add_message.segments(dev_key, seq=0x3129AB, iv_index=0x12345678)
    )

    assert segments[0] == bytes.fromhex("8026ac01ee9dddfd2169326d23f3afdf")
    assert segments[1] == bytes.fromhex("8026ac21cfdc18c52fdef772e0e17308")


def test_application_pdu_segmented_long_mic(dev_key):
    bogus_long_message = AccessMessage(
        src=0x0003, dst=0x1201, ttl=0x04, payload=bytes(range(21))
    )

    segments = list(
        bogus_long_message.segments(dev_key, seq=0x3129AB, iv_index=0x12345678)
    )

    assert len(segments) == 3
    assert segments[0] == bytes.fromhex("80a6ac023008177a35ada597ba2ad9a4")
    assert segments[1] == bytes.fromhex("80a6ac227785110a8afb1ae2fcd561fd")
    assert segments[2] == bytes.fromhex("80a6ac424e48e5be9e")


def test_application_pack_to_network_pdu(
    health_current_status_message: AccessMessage,
    app_key: ApplicationKey,
    net_key: NetworkKey,
):
    network_message = NetworkMessage(health_current_status_message)
    ((seq, network_pdu),) = network_message.pack(
        app_key, net_key, seq=0x000007, iv_index=0x12345678
    )

    assert network_pdu.hex() == "6848cba437860e5673728a627fb938535508e21a6baf57"


def test_application_pack_to_network_pdu_skip_segments(
    app_sar_message: AccessMessage, app_key: ApplicationKey, net_key: NetworkKey,
):
    network_message = NetworkMessage(app_sar_message)
    ((seq, network_pdu),) = network_message.pack(
        app_key,
        net_key,
        seq=0x000007,
        iv_index=0x12345678,
        transport_seq=0x000008,
        skip_segments=[0],
    )

    assert network_pdu.hex() == "689353e01b0a211f82734ad0c7c92081616414e0f8b7e7048c7e"


def test_application_unpack_from_network_pdu(
    health_current_status_message: AccessMessage,
    app_key: ApplicationKey,
    net_key: NetworkKey,
):
    _, _, unpacked_network_message = NetworkMessage.unpack(
        app_key,
        net_key,
        0x12345678,
        bytes.fromhex("6848cba437860e5673728a627fb938535508e21a6baf57"),
    )

    assert unpacked_network_message.message == health_current_status_message


def test_application_network_pdu_segmented_retry(
    health_current_status_message, app_key, net_key
):
    network_message = NetworkMessage(health_current_status_message)

    ((seq, network_pdu),) = network_message.pack(
        app_key, net_key, seq=0x000017, iv_index=0x12345678, transport_seq=0x000007
    )

    assert seq == 0x000017
    assert network_pdu.hex() == "6860f30170e2192e84fb4385254e9e71657aa1bcf2ca90"


def test_control_segment_ack_message(control_appkey_add_ack_message, app_key, net_key):
    network_message = NetworkMessage(control_appkey_add_ack_message)

    ((seq, network_pdu),) = network_message.pack(
        app_key, net_key, seq=0x014835, iv_index=0x12345678
    )

    assert network_pdu.hex() == "68e476b5579c980d0d730f94d7f3509df987bb417eb7c05f"


def test_control_pack_to_network_pdu(control_friend_offer_message, app_key, net_key):
    network_message = NetworkMessage(control_friend_offer_message)

    ((seq, network_pdu),) = network_message.pack(
        app_key, net_key, seq=0x014820, iv_index=0x12345678
    )

    assert network_pdu.hex() == "68d4c826296d7979d7dbc0c9b4d43eebec129d20a620d01e"


def test_control_unpack_from_network_pdu(
    control_friend_offer_message: ControlMessage, net_key: NetworkKey
):
    _, _, unpacked_network_message = NetworkMessage.unpack(
        app_key,
        net_key,
        0x12345678,
        bytes.fromhex("68d4c826296d7979d7dbc0c9b4d43eebec129d20a620d01e"),
    )

    assert unpacked_network_message.message == control_friend_offer_message


def test_proxy_config_pack_to_network_pdu(
    proxy_use_whitelist_message: ProxyConfigMessage,
):
    network_message = NetworkMessage(proxy_use_whitelist_message)
    net_key_local = NetworkKey(bytes.fromhex("d1aafb2a1a3c281cbdb0e960edfad852"))
    ((seq, network_pdu),) = network_message.pack(
        app_key, net_key_local, seq=0x000001, iv_index=0x12345678
    )

    assert network_pdu.hex() == "10386bd60efbbb8b8c28512e792d3711f4b526"


def test_proxy_config_unpack_from_network_pdu(
    proxy_use_whitelist_message: ProxyConfigMessage,
):
    net_key_local = NetworkKey(bytes.fromhex("d1aafb2a1a3c281cbdb0e960edfad852"))
    _, _, unpacked_network_message = NetworkMessage.unpack(
        app_key,
        net_key_local,
        0x12345678,
        bytes.fromhex("10386bd60efbbb8b8c28512e792d3711f4b526"),
        proxy=True,
    )

    assert unpacked_network_message.message == proxy_use_whitelist_message


def test_proxy_solicitation_pack_to_network_pdu(
    proxy_solicitation_message: SolicitationMessage,
):
    network_message = NetworkMessage(proxy_solicitation_message)
    net_key_local = NetworkKey(bytes.fromhex("18eed9c2a56add85049ffc3c59ad0e12"))
    ((seq, network_pdu),) = network_message.pack(
        app_key, net_key_local, seq=0x000001, iv_index=0x00000000
    )

    assert network_pdu.hex() == "7415fd26d31ba53425f13b423508c0019a"
